import zoneinfo

from .log import ActivityLog
from .settings import GlobalSettings

TIMEZONE_CHOICES = [
    tz for tz in zoneinfo.available_timezones() if not tz.startswith("Etc/")
]


__all__ = ["ActivityLog", "GlobalSettings"]



import urllib

from django.contrib import messages
from django.db.models import Q
from django.http import JsonResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.translation import gettext as _
from django.views.generic import View
from django_scopes import scopes_disabled

from pretalx.event.models import Organiser, Team
from pretalx.person.models import SpeakerProfile, User


class UserList(View):
    def dispatch(self, request, *args, **kwargs):
        search = request.GET.get("search")
        if not search or len(search) < 3:
            return JsonResponse({"count": 0, "results": []})

        if request.GET.get("orga", "false").lower() == "true":
            if request.user.is_administrator:
                organisers = Organiser.objects.all()
            else:
                user_teams = request.user.teams.filter(
                    can_change_organiser_settings=True
                )
                organisers = Organiser.objects.filter(teams__in=user_teams)
            teams = Team.objects.filter(organiser__in=organisers)
            users = User.objects.filter(
                Q(name__icontains=search) | Q(email__icontains=search),
                teams__in=teams,
            )[:8]
        else:
            events = self.request.user.get_events_for_permission(
                can_change_submissions=True
            )
            with scopes_disabled():
                users = list(
                    set(
                        SpeakerProfile.objects.filter(
                            Q(user__name__icontains=search)
                            | Q(user__email__icontains=search),
                            event__in=events,
                        ).values_list("user", flat=True)
                    )
                )
            users = User.objects.filter(pk__in=users[:8])

        return JsonResponse(
            {
                "count": len(users),
                "results": [{"email": user.email, "name": user.name} for user in users],
            }
        )


class SubuserView(View):
    def dispatch(self, request, *args, **kwargs):
        request.user.is_administrator = request.user.is_superuser
        request.user.is_superuser = False
        request.user.save(update_fields=["is_administrator", "is_superuser"])
        messages.success(
            request, _("You are now an administrator instead of a superuser.")
        )
        params = request.GET.copy()
        url = urllib.parse.unquote(params.pop("next", [""])[0])
        if url and url_has_allowed_host_and_scheme(url, allowed_hosts=None):
            return redirect(url + ("?" + params.urlencode() if params else ""))
        return redirect(reverse("orga:event.list"))



import io
from collections import defaultdict
from urllib.parse import urlparse

import vobject
from django.conf import settings
from django.core.files.storage import Storage
from django.http import FileResponse, Http404, HttpResponse
from django.shortcuts import redirect
from django.utils.functional import cached_property
from django.views.decorators.cache import cache_page
from django.views.generic import DetailView, ListView, TemplateView
from django_context_decorator import context

from pretalx.common.mixins.views import (
    EventPermissionRequired,
    Filterable,
    PermissionRequired,
    SocialMediaCardMixin,
)
from pretalx.common.utils import safe_filename
from pretalx.person.models import SpeakerProfile, User
from pretalx.submission.models import QuestionTarget


class SpeakerList(EventPermissionRequired, Filterable, ListView):
    context_object_name = "speakers"
    template_name = "agenda/speakers.html"
    permission_required = "agenda.view_schedule"
    default_filters = ("user__name__icontains",)

    def get_queryset(self):
        qs = (
            SpeakerProfile.objects.filter(
                user__in=self.request.event.speakers, event=self.request.event
            )
            .select_related("user", "event")
            .order_by("user__name")
        )
        qs = self.filter_queryset(qs)

        speaker_mapping = defaultdict(list)
        for talk in self.request.event.talks.all().prefetch_related("speakers"):
            for speaker in talk.speakers.all():
                speaker_mapping[speaker.code].append(talk)

        for profile in qs:
            profile.talks = speaker_mapping[profile.user.code]
        return qs


class SpeakerView(PermissionRequired, TemplateView):
    template_name = "agenda/speaker.html"
    permission_required = "agenda.view_speaker"
    slug_field = "code"

    @context
    @cached_property
    def profile(self):
        return (
            SpeakerProfile.objects.filter(
                event=self.request.event, user__code__iexact=self.kwargs["code"]
            )
            .select_related("user")
            .first()
        )

    @context
    @cached_property
    def talks(self):
        if not self.request.event.current_schedule:
            return []
        return self.request.event.current_schedule.talks.filter(
            submission__speakers__code=self.kwargs["code"], is_visible=True
        )

    def get_permission_object(self):
        return self.profile

    @context
    def answers(self):
        return self.profile.user.answers.filter(
            question__is_public=True,
            question__event=self.request.event,
            question__target=QuestionTarget.SPEAKER,
        ).select_related("question")


class SpeakerRedirect(DetailView):
    model = User

    def dispatch(self, request, **kwargs):
        speaker = self.get_object()
        profile = speaker.profiles.filter(event=self.request.event).first()
        if profile and self.request.user.has_perm("agenda.view_speaker", profile):
            return redirect(profile.urls.public.full())
        raise Http404()


class SpeakerTalksIcalView(PermissionRequired, DetailView):
    context_object_name = "profile"
    permission_required = "agenda.view_speaker"
    slug_field = "code"

    def get_object(self, queryset=None):
        return SpeakerProfile.objects.filter(
            event=self.request.event, user__code__iexact=self.kwargs["code"]
        ).first()

    def get(self, request, event, *args, **kwargs):
        netloc = urlparse(settings.SITE_URL).netloc
        speaker = self.get_object()
        slots = self.request.event.current_schedule.talks.filter(
            submission__speakers=speaker.user, is_visible=True
        ).select_related("room", "submission")

        cal = vobject.iCalendar()
        cal.add(
            "prodid"
        ).value = f"-//pretalx//{netloc}//{request.event.slug}//{speaker.code}"

        for slot in slots:
            slot.build_ical(cal)

        speaker_name = Storage().get_valid_name(name=speaker.user.name)
        return HttpResponse(
            cal.serialize(),
            content_type="text/calendar",
            headers={
                "Content-Disposition": f'attachment; filename="{request.event.slug}-{safe_filename(speaker_name)}.ics"'
            },
        )


class SpeakerSocialMediaCard(SocialMediaCardMixin, SpeakerView):
    def get_image(self):
        user = self.profile.user
        if user.avatar:
            return user.avatar


@cache_page(60 * 60)
def empty_avatar_view(request, event):
    # cached for an hour
    color = request.event.primary_color or "#3aa57c"
    avatar_template = f"""<svg
   xmlns="http://www.w3.org/2000/svg"
   viewBox="0 0 100 100">
  <g>
    <path
       id="body"
       d="m 2,98 h 96 0 c 0,0 6,-65 -48,-52 c 0,0 -54,-10 -48,52"
       style="fill:none;stroke:{color};stroke-width:1.6;stroke-linecap:butt;stroke-linejoin:round;stroke-miterlimit:4;stroke-dasharray:2.1, 2.1;stroke-dashoffset:0;stroke-opacity:0.87" />
    <ellipse
       ry="27"
       rx="27"
       cy="28"
       cx="50"
       id="heady"
       style="fill:#ffffff;stroke:{color};stroke-width:1.3;stroke-linecap:butt;stroke-linejoin:round;stroke-miterlimit:4;stroke-dasharray:6.5, 8;stroke-dashoffset:4;stroke-opacity:0.87" />
  </g>
</svg>"""
    return FileResponse(
        io.BytesIO(avatar_template.encode()),
        as_attachment=True,
        content_type="image/svg+xml",
    )



import pytest

from pretalx.agenda.recording import BaseRecordingProvider


def test_provider_raises_not_implemented():
    with pytest.raises(NotImplementedError):
        BaseRecordingProvider("event").get_recording("submission")



# Generated by Django 3.2.16 on 2022-11-16 22:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mail", "0009_queuedmail_locale"),
    ]

    operations = [
        migrations.AddField(
            model_name="queuedmail",
            name="attachments",
            field=models.JSONField(default=None, null=True),
        ),
    ]



import pytest
from django_scopes import scope

from .utils import screenshot


@pytest.mark.django_db
def screenshot_agenda_public_schedule(live_server, client, event):
    with scope(event=event):
        client.get(live_server.url + f"/{event.slug}/schedule/")
    screenshot(client, "website/agenda_public.png")


@pytest.mark.django_db
def screenshot_cfp_submission_info(live_server, client, event):
    with scope(event=event):
        client.get(live_server.url + f"/{event.slug}/submit/423mOn/info/")
    screenshot(client, "website/cfp_start.png")


@pytest.mark.django_db
def screenshot_cfp_submission_questions(live_server, client, event):
    with scope(event=event):
        client.get(live_server.url + f"/{event.slug}/submit/423mOn/questions/")
    screenshot(client, "website/cfp_questions.png")


@pytest.mark.django_db
def screenshot_edit_cfp_settings(live_server, logged_in_client, event):
    with scope(event=event):
        logged_in_client.get(
            live_server.url + f"/orga/event/{event.slug}/cfp/text#information"
        )
    screenshot(logged_in_client, "website/cfp_settings.png")


@pytest.mark.django_db
def screenshot_edit_question_settings(live_server, logged_in_client, event):
    with scope(event=event):
        logged_in_client.get(
            live_server.url + f"/orga/event/{event.slug}/cfp/questions/new"
        )
        logged_in_client.find_element_by_css_selector("#id_variant").click()
    screenshot(logged_in_client, "website/question_settings.png")


@pytest.mark.django_db
def screenshot_edit_plugins(live_server, logged_in_client, user, event):
    with scope(event=event):
        user.is_administrator = True
        user.save()
        logged_in_client.get(
            live_server.url + f"/orga/event/{event.slug}/settings/plugins"
        )
    screenshot(logged_in_client, "website/plugin_settings.png", scroll=False)


@pytest.mark.django_db
def screenshot_edit_mail_templates(live_server, logged_in_client, event):
    with scope(event=event):
        logged_in_client.get(
            live_server.url + f"/orga/event/{event.slug}/mails/templates"
        )
    screenshot(logged_in_client, "website/mail_templates.png")


@pytest.mark.django_db
def screenshot_review_submission(live_server, logged_in_client, event):
    with scope(event=event):
        submission = event.submissions.first()
        logged_in_client.get(
            live_server.url
            + "/orga/event/{}/submissions/{}/reviews".format(
                event.slug, submission.code
            )
        )
    screenshot(logged_in_client, "website/review_submission.png")


@pytest.mark.django_db
def screenshot_bare_schedule_editor(live_server, logged_in_client, event):
    with scope(event=event):
        logged_in_client.get(live_server.url + f"/orga/event/{event.slug}/schedule/")
        logged_in_client.execute_script(
            """
    const selectors = [".alert", ".schedule-header"]
    for (selector of selectors) {
        var element = document.querySelector(selector);
        if (element)
            element.parentNode.removeChild(element);
    }"""
        )
    screenshot(logged_in_client, "website/edit_schedule.png")


@pytest.mark.django_db
def screenshot_export_schedule_editor(live_server, logged_in_client, event):
    with scope(event=event):
        event.wip_schedule.freeze("v1")
        logged_in_client.get(
            live_server.url + f"/orga/event/{event.slug}/schedule/export"
        )
        logged_in_client.find_element_by_css_selector("#custom-tab").click()
    screenshot(logged_in_client, "website/schedule_export.png")



import pytest
from django_scopes import scope


@pytest.mark.django_db
def test_log_hides_password(submission):
    with scope(event=submission.event):
        submission.log_action(
            "test.hide", data={"password": "12345", "non-sensitive": "foo"}
        )
        log = submission.logged_actions().get(action_type="test.hide")
        assert log.json_data["password"] != "12345"
        assert log.json_data["non-sensitive"] == "foo"


@pytest.mark.django_db
def test_log_wrong_data(submission):
    with scope(event=submission.event):
        with pytest.raises(TypeError):
            submission.log_action(
                "test.hide", data=[{"password": "12345", "non-sensitive": "foo"}]
            )



import datetime as dt
import re
import string
import uuid
from contextlib import suppress
from urllib.parse import urlparse
from zoneinfo import ZoneInfo

import vobject
from django.conf import settings
from django.db import models
from django.utils.functional import cached_property
from django_scopes import ScopedManager
from i18nfield.fields import I18nCharField

from pretalx.common.mixins.models import LogMixin
from pretalx.common.urls import get_base_url

INSTANCE_IDENTIFIER = None
with suppress(Exception):
    from pretalx.common.models.settings import GlobalSettings

    INSTANCE_IDENTIFIER = GlobalSettings().get_instance_identifier()


class TalkSlot(LogMixin, models.Model):
    """The TalkSlot object is the scheduled version of a.

    :class:`~pretalx.submission.models.submission.Submission`.

    TalkSlots always belong to one submission and one :class:`~pretalx.schedule.models.schedule.Schedule`.

    :param is_visible: This parameter is set on schedule release. Only confirmed talks will be visible.
    """

    submission = models.ForeignKey(
        to="submission.Submission",
        on_delete=models.PROTECT,
        related_name="slots",
        null=True,
        blank=True,  # If the submission is empty, this is a break or similar event
    )
    room = models.ForeignKey(
        to="schedule.Room",
        on_delete=models.PROTECT,
        related_name="talks",
        null=True,
        blank=True,
    )
    schedule = models.ForeignKey(
        to="schedule.Schedule", on_delete=models.PROTECT, related_name="talks"
    )
    is_visible = models.BooleanField(default=False)
    start = models.DateTimeField(null=True)
    end = models.DateTimeField(null=True)
    description = I18nCharField(null=True)

    updated = models.DateTimeField(auto_now=True)

    objects = ScopedManager(event="schedule__event")

    class Meta:
        ordering = ("start",)

    def __str__(self):
        """Help when debugging."""
        return f'TalkSlot(event={self.schedule.event.slug}, submission={getattr(self.submission, "title", None)}, schedule={self.schedule.version})'

    @cached_property
    def event(self):
        return self.submission.event if self.submission else self.schedule.event

    @property
    def duration(self) -> int:
        """Returns the actual duration in minutes if the talk is scheduled, and
        the planned duration in minutes otherwise."""
        if self.start and self.end:
            return int((self.end - self.start).total_seconds() / 60)
        if not self.submission:
            return None
        return self.submission.get_duration()

    @cached_property
    def export_duration(self):
        from pretalx.common.serialize import serialize_duration

        return serialize_duration(minutes=self.duration)

    @cached_property
    def pentabarf_export_duration(self):
        duration = dt.timedelta(minutes=self.duration)
        days = duration.days
        hours = duration.total_seconds() // 3600 - days * 24
        minutes = duration.seconds // 60 % 60
        return f"{hours:02}{minutes:02}00"

    @cached_property
    def local_start(self):
        if self.start:
            return self.start.astimezone(self.event.tz)

    @cached_property
    def real_end(self):
        """Guaranteed to provide a useful end datetime if ``start`` is set,
        even if ``end`` is empty."""
        return self.end or (
            self.start + dt.timedelta(minutes=self.duration) if self.start else None
        )

    @cached_property
    def local_end(self):
        if self.real_end:
            return self.real_end.astimezone(self.event.tz)

    @cached_property
    def as_availability(self):
        """'Casts' a slot as.

        :class:`~pretalx.schedule.models.availability.Availability`, useful for
        availability arithmetic.
        """
        from pretalx.schedule.models import Availability

        return Availability(
            start=self.start,
            end=self.real_end,
        )

    def copy_to_schedule(self, new_schedule, save=True):
        """Create a new slot for the given.

        :class:`~pretalx.schedule.models.schedule.Schedule` with all other
        fields identical to this one.
        """
        new_slot = TalkSlot(schedule=new_schedule)

        for field in [f for f in self._meta.fields if f.name not in ("id", "schedule")]:
            setattr(new_slot, field.name, getattr(self, field.name))

        if save:
            new_slot.save()
        return new_slot

    copy_to_schedule.alters_data = True

    def is_same_slot(self, other_slot) -> bool:
        """Checks if both slots have the same room and start time."""
        return self.room == other_slot.room and self.start == other_slot.start

    @cached_property
    def id_suffix(self):
        if not self.event.feature_flags["present_multiple_times"]:
            return ""
        all_slots = list(
            TalkSlot.objects.filter(
                submission_id=self.submission_id, schedule_id=self.schedule_id
            )
        )
        if len(all_slots) == 1:
            return ""
        return "-" + str(all_slots.index(self))

    @cached_property
    def frab_slug(self):
        title = re.sub(r"\W+", "-", self.submission.title)
        legal_chars = string.ascii_letters + string.digits + "-"
        pattern = f"[^{legal_chars}]+"
        title = re.sub(pattern, "", title)
        title = title.lower()
        title = title.strip("_")
        return f"{self.event.slug}-{self.submission.pk}{self.id_suffix}-{title}"

    @cached_property
    def uuid(self):
        """A UUID5, calculated from the submission code and the instance
        identifier."""
        global INSTANCE_IDENTIFIER
        if not INSTANCE_IDENTIFIER:
            from pretalx.common.models.settings import GlobalSettings

            INSTANCE_IDENTIFIER = GlobalSettings().get_instance_identifier()
        return uuid.uuid5(INSTANCE_IDENTIFIER, self.submission.code + self.id_suffix)

    def build_ical(self, calendar, creation_time=None, netloc=None):
        if not self.start or not self.local_end or not self.room or not self.submission:
            return
        creation_time = creation_time or dt.datetime.now(ZoneInfo("UTC"))
        netloc = netloc or urlparse(get_base_url(self.event)).netloc

        vevent = calendar.add("vevent")
        vevent.add(
            "summary"
        ).value = f"{self.submission.title} - {self.submission.display_speaker_names}"
        vevent.add("dtstamp").value = creation_time
        vevent.add("location").value = str(self.room.name)
        vevent.add("uid").value = "pretalx-{}-{}{}@{}".format(
            self.submission.event.slug, self.submission.code, self.id_suffix, netloc
        )

        vevent.add("dtstart").value = self.local_start
        vevent.add("dtend").value = self.local_end
        vevent.add("description").value = self.submission.abstract or ""
        vevent.add("url").value = self.submission.urls.public.full()

    def full_ical(self):
        netloc = urlparse(settings.SITE_URL).netloc
        cal = vobject.iCalendar()
        cal.add("prodid").value = "-//pretalx//{}//{}".format(
            netloc, self.submission.code
        )
        self.build_ical(cal)
        return cal



from django import forms
from django.db.models import Count
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django_scopes.forms import SafeModelChoiceField

from pretalx.cfp.forms.cfp import CfPFormMixin
from pretalx.common.forms.fields import ImageField
from pretalx.common.forms.widgets import MarkdownWidget
from pretalx.common.mixins.forms import PublicContent, RequestRequire
from pretalx.submission.forms.track_select_widget import TrackSelectWidget
from pretalx.submission.models import Question, Submission, SubmissionStates


class InfoForm(CfPFormMixin, RequestRequire, PublicContent, forms.ModelForm):
    additional_speaker = forms.EmailField(
        label=_("Additional Speaker"),
        help_text=_(
            "If you have a co-speaker, please add their email address here, and we will invite them to create an account. If you have more than one co-speaker, you can add more speakers after finishing the proposal process."
        ),
        required=False,
    )
    image = ImageField(
        required=False,
        label=_("Session image"),
        help_text=_("Use this if you want an illustration to go with your proposal."),
    )
    content_locale = forms.ChoiceField(label=_("Language"))

    def __init__(self, event, **kwargs):
        self.event = event
        self.readonly = kwargs.pop("readonly", False)
        self.access_code = kwargs.pop("access_code", None)
        self.default_values = {}
        instance = kwargs.get("instance")
        initial = kwargs.pop("initial", {}) or {}
        if not instance or not instance.submission_type:
            initial["submission_type"] = (
                getattr(self.access_code, "submission_type", None)
                or initial.get("submission_type")
                or self.event.cfp.default_type
            )
        if not instance and self.access_code:
            initial["track"] = self.access_code.track
        if not instance or not instance.content_locale:
            initial["content_locale"] = self.event.locale

        super().__init__(initial=initial, **kwargs)

        if "abstract" in self.fields:
            self.fields["abstract"].widget.attrs["rows"] = 2

        self._set_track(instance=instance)
        self._set_submission_types(instance=instance)
        self._set_locales()
        self._set_slot_count(instance=instance)

        if self.readonly:
            for f in self.fields.values():
                f.disabled = True

    def _set_track(self, instance=None):
        if "track" in self.fields:
            if (
                not self.event.feature_flags["use_tracks"]
                or instance
                and instance.state != SubmissionStates.SUBMITTED
            ):
                self.fields.pop("track")
                return

            access_code = self.access_code or getattr(instance, "access_code", None)
            if not access_code or not access_code.track:
                self.fields["track"].queryset = self.event.tracks.filter(
                    requires_access_code=False
                )
            else:
                self.fields["track"].queryset = self.event.tracks.filter(
                    pk=access_code.track.pk
                )
            if (
                len(self.fields["track"].queryset) == 1
                and self.fields["track"].required
            ):
                self.default_values["track"] = self.fields["track"].queryset.first()
                self.fields.pop("track")

    def _set_submission_types(self, instance=None):
        _now = now()
        submission_types = self.event.submission_types
        if (
            instance
            and instance.pk
            and (
                instance.state != SubmissionStates.SUBMITTED
                or not self.event.cfp.is_open
            )
        ):
            self.fields["submission_type"].queryset = submission_types.filter(
                pk=instance.submission_type_id
            )
            self.fields["submission_type"].disabled = True
            return
        access_code = self.access_code or getattr(instance, "access_code", None)
        if access_code and not access_code.submission_type:
            pks = set(submission_types.values_list("pk", flat=True))
        elif access_code:
            pks = {access_code.submission_type.pk}
        else:
            queryset = submission_types.filter(requires_access_code=False)
            if not self.event.cfp.deadline or self.event.cfp.deadline >= _now:
                # No global deadline or still open
                types = queryset.exclude(deadline__lt=_now)
            else:
                types = queryset.filter(deadline__gte=_now)
            pks = set(types.values_list("pk", flat=True))
        if instance and instance.pk:
            pks |= {instance.submission_type.pk}
        if len(pks) == 1:
            self.default_values["submission_type"] = submission_types.get(
                pk=list(pks)[0]
            )
            self.fields.pop("submission_type")
        else:
            self.fields["submission_type"].queryset = submission_types.filter(
                pk__in=pks
            )
            if "duration" in self.fields and not self.fields["duration"].required:
                self.fields["duration"].help_text += " " + str(
                    _("Leave empty to use the default duration for the session type.")
                )

    def _set_locales(self):
        if "content_locale" in self.fields:
            if len(self.event.content_locales) == 1:
                self.default_values["content_locale"] = self.event.content_locales[0]
                self.fields.pop("content_locale")
            else:
                self.fields["content_locale"].choices = self.event.named_content_locales

    def _set_slot_count(self, instance=None):
        if not self.event.feature_flags["present_multiple_times"]:
            self.fields.pop("slot_count", None)
        elif (
            "slot_count" in self.fields
            and instance
            and instance.state
            in [SubmissionStates.ACCEPTED, SubmissionStates.CONFIRMED]
        ):
            self.fields["slot_count"].disabled = True
            self.fields["slot_count"].help_text += " " + str(
                _(
                    "Please contact the organisers if you want to change how often you're presenting this proposal."
                )
            )

    def save(self, *args, **kwargs):
        for key, value in self.default_values.items():
            setattr(self.instance, key, value)
        return super().save(*args, **kwargs)

    class Meta:
        model = Submission
        fields = [
            "title",
            "submission_type",
            "track",
            "content_locale",
            "abstract",
            "description",
            "notes",
            "slot_count",
            "do_not_record",
            "image",
            "duration",
        ]
        request_require = [
            "title",
            "abstract",
            "description",
            "notes",
            "image",
            "do_not_record",
            "track",
            "duration",
            "content_locale",
            "additional_speaker",
        ]
        public_fields = ["title", "abstract", "description", "image"]
        widgets = {
            "abstract": MarkdownWidget,
            "description": MarkdownWidget,
            "notes": MarkdownWidget,
            "track": TrackSelectWidget,
        }
        field_classes = {
            "submission_type": SafeModelChoiceField,
            "track": SafeModelChoiceField,
        }


class SelectMultipleWithCount(forms.SelectMultiple):
    """A widget for multi-selects that correspond to countable values.

    This widget doesn't support some of the options of the default
    SelectMultiple, most notably it doesn't support optgroups. In
    return, it takes a third value per choice, makes zero-values
    disabled and sorts options by numerical value.
    """

    def optgroups(self, name, value, attrs=None):
        choices = sorted(self.choices, key=lambda x: x[1].count, reverse=True)
        result = []
        for index, (option_value, label) in enumerate(choices):
            selected = str(option_value) in value
            result.append(
                self.create_option(
                    name,
                    value=option_value,
                    label=label,
                    selected=selected,
                    index=index,
                )
            )
        return [(None, result, 0)]

    def create_option(self, name, value, label, *args, count=0, **kwargs):
        option = super().create_option(name, value, str(label), *args, **kwargs)
        if label.count == 0:
            option["attrs"]["class"] = "hidden"
        return option


class CountableOption:
    def __init__(self, name, count):
        self.name = name
        self.count = count

    def __str__(self):
        return f"{self.name} ({self.count})"


class SubmissionFilterForm(forms.Form):
    state = forms.MultipleChoiceField(
        required=False,
        widget=SelectMultipleWithCount(
            attrs={"class": "select2", "title": _("Proposal states")}
        ),
        choices=SubmissionStates.get_choices(),
    )
    submission_type = forms.MultipleChoiceField(
        required=False,
        widget=SelectMultipleWithCount(
            attrs={"class": "select2", "title": _("Session types")}
        ),
    )
    pending_state__isnull = forms.BooleanField(
        required=False,
        label=_("exclude pending"),
    )
    content_locale = forms.MultipleChoiceField(
        required=False,
        widget=SelectMultipleWithCount(
            attrs={"class": "select2", "title": _("Language")}
        ),
    )
    track = forms.MultipleChoiceField(
        required=False,
        widget=SelectMultipleWithCount(
            attrs={"class": "select2", "title": _("Tracks")}
        ),
    )
    tags = forms.MultipleChoiceField(
        required=False,
        widget=SelectMultipleWithCount(attrs={"class": "select2", "title": _("Tags")}),
    )
    question = SafeModelChoiceField(queryset=Question.objects.none(), required=False)

    def __init__(self, event, *args, limit_tracks=False, **kwargs):
        self.event = event
        usable_states = kwargs.pop("usable_states", None)
        initial = kwargs.pop("initial", {}) or {}
        initial["exclude_pending"] = False
        super().__init__(*args, initial=initial, **kwargs)
        qs = event.submissions
        state_qs = Submission.objects.filter(event=event)
        if usable_states:
            qs = qs.filter(state__in=usable_states)
            state_qs = state_qs.filter(state__in=usable_states)
        state_count = {
            d["state"]: d["state__count"]
            for d in state_qs.order_by("state").values("state").annotate(Count("state"))
        }
        state_count.update(
            {
                f"pending_state__{d['pending_state']}": d["pending_state__count"]
                for d in state_qs.filter(pending_state__isnull=False)
                .order_by("pending_state")
                .values("pending_state")
                .annotate(Count("pending_state"))
            }
        )
        sub_types = event.submission_types.all()
        tracks = limit_tracks or event.tracks.all()
        languages = event.named_content_locales
        if len(sub_types) > 1:
            type_count = {
                d["submission_type_id"]: d["submission_type_id__count"]
                for d in qs.order_by("submission_type_id")
                .values("submission_type_id")
                .annotate(Count("submission_type_id"))
            }
            self.fields["submission_type"].choices = [
                (
                    sub_type.pk,
                    CountableOption(sub_type.name, type_count.get(sub_type.pk, 0)),
                )
                for sub_type in sub_types
            ]
        else:
            self.fields.pop("submission_type", None)
        if len(tracks) > 1:
            track_count = {
                d["track"]: d["track__count"]
                for d in qs.order_by("track").values("track").annotate(Count("track"))
            }
            self.fields["track"].choices = [
                (track.pk, CountableOption(track.name, track_count.get(track.pk, 0)))
                for track in tracks
            ]
        else:
            self.fields.pop("track", None)
        if len(languages) > 1:
            language_count = {
                d["content_locale"]: d["content_locale__count"]
                for d in qs.order_by("content_locale")
                .values("content_locale")
                .annotate(Count("content_locale"))
            }
            self.fields["content_locale"].choices = [
                (code, CountableOption(name, language_count.get(code, 0)))
                for code, name in languages
            ]
        else:
            self.fields.pop("content_locale", None)

        if not self.event.tags.all().exists():
            self.fields.pop("tags", None)
        else:
            tag_count = event.tags.prefetch_related("submissions").annotate(
                submission_count=Count("submissions", distinct=True)
            )
            tag_count = {tag.tag: tag.submission_count for tag in tag_count}
            self.fields["tags"].choices = [
                (tag.pk, CountableOption(tag.tag, tag_count.get(tag.tag, 0)))
                for tag in self.event.tags.all()
            ]

        if usable_states:
            usable_states = [
                choice
                for choice in self.fields["state"].choices
                if choice[0] in usable_states
            ]
        else:
            usable_states = self.fields["state"].choices
        usable_states += [
            (
                f"pending_state__{choice[0]}",
                str(_("Pending {state}")).format(state=choice[1]),
            )
            for choice in usable_states
        ]
        self.fields["state"].choices = [
            (
                choice[0],
                CountableOption(choice[1].capitalize(), state_count.get(choice[0], 0)),
            )
            for choice in usable_states
        ]
        self.fields["question"].queryset = event.questions.all()



from django.db import transaction
from django_scopes import scopes_disabled

from pretalx.event.models import Event
from pretalx.submission.models import Submission


@scopes_disabled()
@transaction.atomic
def move_submission(code, new_event, copy=False):
    """Caution! Does not include.

    - submission type mapping (resets to default)
    - questions with options (choice/multiple choice)
    - other questions only if they are an exact string match
    - tags
    - tracks
    - review scores, presumably

    Set copy=True to not delete the proposal in the old event.
    """
    submission = Submission.objects.get(code__iexact=code)
    old_event = submission.event
    new_event = Event.objects.get(slug__iexact=new_event)

    speaker_questions = {
        str(q.question): q for q in new_event.questions.all().filter(target="speaker")
    }
    submission_questions = {
        str(q.question): q
        for q in new_event.questions.all().filter(target="submission")
    }

    if copy:
        submission.id = None
        submission.code += "C"
        submission.review_code = None
    submission.event = new_event
    submission.submission_type = new_event.cfp.default_type
    submission.save()

    for answer in (
        submission.answers.all()
        .filter(question__event=old_event)
        .select_related("question")
    ):
        other_question = submission_questions.get(str(answer.question.question))
        if other_question:
            answer.id = None
            answer.question = other_question
            answer.save()

    for speaker in submission.speakers.all():
        if not speaker.profiles.filter(event=new_event).exists():
            old_profile = speaker.event_profile(old_event)
            old_profile.pk = None
            old_profile.event = new_event
            old_profile.save()
        for answer in (
            speaker.answers.all()
            .filter(question__event=old_event)
            .select_related("question")
        ):
            other_question = speaker_questions.get(str(answer.question.question))
            if other_question:
                answer.id = None
                answer.question = other_question
                answer.save()



# Generated by Django 1.11.3 on 2017-07-24 11:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mail", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="queuedmail",
            name="sent",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



import string
from collections import defaultdict
from contextlib import suppress

from django import forms
from django.db import transaction
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from i18nfield.forms import I18nModelForm

from pretalx.common.exceptions import SendMailException
from pretalx.common.mixins.forms import I18nHelpText, ReadOnlyFlag
from pretalx.mail.context import get_available_placeholders
from pretalx.mail.models import MailTemplate, QueuedMail
from pretalx.person.models import User
from pretalx.submission.models.submission import Submission, SubmissionStates


class MailTemplateBase(I18nHelpText, I18nModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def display_placeholder(key, placeholder):
            result = "{" + f"{key}" + "}"
            if not placeholder or not placeholder.explanation:
                return result
            title = placeholder.explanation
            sample = placeholder.render_sample(self.event)
            if sample:
                title += f"; '{sample}'"
            return f'<span data-toggle="tooltip" title="{title}">{result}</span>'

        available_placeholders = ", ".join(
            [
                display_placeholder(key, placeholder)
                for key, placeholder in self.get_valid_placeholders().items()
                if (not placeholder or placeholder.is_visible)
            ]
        )
        self.fields["text"].help_text = (
            _("Available placeholders:") + " " + available_placeholders
        )
        self.fields["subject"].help_text = (
            _("Available placeholders:") + " " + available_placeholders
        )

    def _clean_for_placeholders(self, text, valid_placeholders):
        cleaned_data = super().clean()
        valid_placeholders = set(valid_placeholders)
        # We need to do this here, since we need to be sure about the valid placeholders first
        used_placeholders = set()
        for field in ("subject", "text"):
            value = cleaned_data.get(field)
            if not value:
                continue
            for lang in value.data.values():
                used_placeholders |= {
                    v[1] for v in string.Formatter().parse(lang) if v[1]
                }
        return used_placeholders - valid_placeholders

    class Meta:
        model = MailTemplate
        fields = ["subject", "text"]


class MailTemplateForm(ReadOnlyFlag, MailTemplateBase):
    def __init__(self, *args, event=None, **kwargs):
        self.event = event
        if event:
            kwargs["locales"] = event.locales
        super().__init__(*args, **kwargs)

    def get_valid_placeholders(self):
        kwargs = ["event", "submission", "user", "slot"]
        valid_placeholders = {}

        if self.instance and self.instance.id:
            if self.instance == self.event.update_template:
                valid_placeholders["notifications"] = []
                kwargs = ["event", "user"]
            elif self.instance == self.event.question_template:
                valid_placeholders["questions"] = []
                valid_placeholders["url"] = []
                kwargs = ["event", "user"]
            elif self.instance in (
                self.event.accept_template,
                self.event.reject_template,
            ):
                kwargs.remove("slot")

        valid_placeholders.update(
            get_available_placeholders(event=self.event, kwargs=kwargs)
        )
        return valid_placeholders

    def clean_text(self):
        text = self.cleaned_data["text"]
        valid_placeholders = self.get_valid_placeholders().keys()
        try:
            warnings = self._clean_for_placeholders(text, valid_placeholders)
        except Exception:
            raise forms.ValidationError(
                _(
                    "Invalid email template! "
                    "Please check that you don't have stray { or } somewhere, "
                    "and that there are no spaces inside the {} blocks."
                )
            )
        if warnings:
            warnings = ", ".join("{" + w + "}" for w in warnings)
            raise forms.ValidationError(
                str(_("Unknown template key!")) + " " + warnings
            )
        return text

    class Meta:
        model = MailTemplate
        fields = ["subject", "text", "reply_to", "bcc"]


class DraftRemindersForm(MailTemplateForm):
    def get_valid_placeholders(self):
        kwargs = ["event", "submission", "user"]
        return get_available_placeholders(event=self.event, kwargs=kwargs)

    def save(self, *args, **kwargs):
        template = self.instance
        submissions = Submission.all_objects.filter(
            state=SubmissionStates.DRAFT, event=self.event
        )
        mail_count = 0
        for submission in submissions:
            for user in submission.speakers.all():
                template.to_mail(
                    user=user,
                    event=self.event,
                    locale=submission.get_email_locale(user.locale),
                    context_kwargs={"submission": submission, "user": user},
                    skip_queue=True,
                    commit=False,
                )
                mail_count += 1

        return mail_count

    class Meta:
        model = MailTemplate
        fields = ["subject", "text"]


class MailDetailForm(ReadOnlyFlag, forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.instance or not self.instance.to_users.all().count():
            self.fields.pop("to_users")
        else:
            self.fields["to_users"].queryset = self.instance.to_users.all()
            self.fields["to_users"].required = False

    def clean(self, *args, **kwargs):
        cleaned_data = super().clean(*args, **kwargs)
        if not cleaned_data["to"] and not cleaned_data.get("to_users"):
            self.add_error(
                "to",
                forms.ValidationError(
                    _("An email needs to have at least one recipient.")
                ),
            )
        return cleaned_data

    def save(self, *args, **kwargs):
        obj = super().save(*args, **kwargs)
        if self.has_changed() and "to" in self.changed_data:
            addresses = list(
                {a.strip().lower() for a in (obj.to or "").split(",") if a.strip()}
            )
            for address in addresses:
                user = User.objects.filter(email__iexact=address).first()
                if user:
                    addresses.remove(address)
                    obj.to_users.add(user)
            addresses = ",".join(addresses) if addresses else ""
            obj.to = addresses
            obj.save()
        return obj

    class Meta:
        model = QueuedMail
        fields = ["to", "to_users", "reply_to", "cc", "bcc", "subject", "text"]
        widgets = {"to_users": forms.SelectMultiple(attrs={"class": "select2"})}


class WriteMailForm(MailTemplateBase):
    recipients = forms.MultipleChoiceField(
        label=_("Recipient groups"),
        choices=(
            (
                "submitted",
                _("Everyone with proposal(s) that have not been accepted/rejected yet"),
            ),
            (
                "accepted",
                _("All accepted speakers (who have not confirmed their session yet)"),
            ),
            ("confirmed", _("All confirmed speakers")),
            ("rejected", _("All rejected speakers")),
            ("canceled", _("All canceled speakers")),
            ("reviewers", _("All reviewers in your team")),
            ("no_slides", _("All confirmed speakers who have not uploaded slides")),
        ),
        required=False,
        widget=forms.SelectMultiple(
            attrs={"class": "select2", "title": _("Recipient groups")}
        ),
    )
    tracks = forms.MultipleChoiceField(
        label=_("All proposals in these tracks"),
        required=False,
        widget=forms.SelectMultiple(attrs={"class": "select2", "title": _("Tracks")}),
        help_text=_("Leave empty to include proposals from all tracks."),
    )
    submission_types = forms.MultipleChoiceField(
        label=_("All proposals of these types"),
        required=False,
        widget=forms.SelectMultiple(
            attrs={"class": "select2", "title": _("Session types")}
        ),
        help_text=_("Leave empty to include proposals of all session types."),
    )
    submissions = forms.MultipleChoiceField(
        required=False,
        label=_("Proposals"),
        widget=forms.SelectMultiple(
            attrs={"class": "select2", "title": _("Proposals")}
        ),
        help_text=_(
            "Select proposals that should receive the email regardless of the other filters."
        ),
    )

    def __init__(self, event, **kwargs):
        self.event = event
        if event:
            kwargs["locales"] = event.locales
        super().__init__(**kwargs)
        self.fields["submissions"].choices = [
            (sub.code, sub.title) for sub in event.submissions.all()
        ]
        if event.feature_flags["use_tracks"] and event.tracks.all().exists():
            self.fields["tracks"].choices = [
                (track.pk, track.name) for track in event.tracks.all()
            ]
        else:
            del self.fields["tracks"]
        self.fields["submission_types"].choices = [
            (submission_type.pk, submission_type.name)
            for submission_type in event.submission_types.all()
        ]
        if len(self.event.locales) > 1:
            self.fields["subject"].help_text = _(
                "If you provide only one language, that language will be used for all emails. If you provide multiple languages, the best fit for each speaker will be used."
            )
        self.warnings = []

    def get_valid_placeholders(self):
        kwargs = ["event", "submission", "user"]
        if getattr(self, "cleaned_data", None):
            recipients = self.cleaned_data.get("recipients")
            if recipients and not set(recipients) - {"accepted", "confirmed"}:
                kwargs.append("slot")
        else:  # Not cleaned yet, we only need this for the help text
            kwargs.append("slot")
        return get_available_placeholders(event=self.event, kwargs=kwargs)

    def get_recipient_submissions(self):
        # If no recipient base groups are selected,
        submissions = self.event.submissions.all()
        submission_states = [
            s
            for s in ["submitted", "accepted", "confirmed", "rejected", "canceled"]
            if s in self.cleaned_data["recipients"]
        ]
        if submission_states:
            query = Q(state__in=submission_states)
        elif "no_slides" in self.cleaned_data["recipients"]:
            query = Q(state__isnull=False)
        else:
            query = Q()
        if "no_slides" in self.cleaned_data["recipients"]:
            if submission_states:
                query = query | Q(resources__isnull=True)
            else:
                query = Q(resources__isnull=True)

        submissions = submissions.filter(query)

        tracks = self.cleaned_data.get("tracks")
        if tracks:
            submissions = submissions.filter(track__in=tracks)

        submission_types = self.cleaned_data.get("submission_types")
        if submission_types:
            submissions = submissions.filter(submission_type__in=submission_types)

        submissions = submissions.select_related(
            "track", "submission_type", "event"
        ).prefetch_related("speakers")

        # Specifically filtered-for submissions need to come last, so they can't be excluded
        filter_submissions = self.cleaned_data.get("submissions")
        if filter_submissions:
            filtered = any(
                self.cleaned_data.get(key)
                for key in ["recipients", "tracks", "submission_types"]
            )
            specific_submissions = (
                self.event.submissions.filter(code__in=filter_submissions)
                .select_related("track", "submission_type", "event")
                .prefetch_related("speakers")
            )
            if filtered:
                # If we have filtered on things already, we add our specific sessions to the existing set
                submissions = submissions.union(specific_submissions)
            else:
                # Otherwise, we assume that the user wanted *only* the specific sessions
                submissions = specific_submissions
        return submissions

    def clean(self):
        cleaned_data = super().clean()
        valid_placeholders = self.get_valid_placeholders().keys()
        self.warnings = self._clean_for_placeholders(
            cleaned_data.get("subject", ""), valid_placeholders
        ) | self._clean_for_placeholders(
            cleaned_data.get("text", ""), valid_placeholders
        )
        return cleaned_data

    @transaction.atomic
    def save(self):
        self.instance.event = self.event
        self.instance.is_auto_created = True
        template = super().save()

        submissions = self.get_recipient_submissions()
        mails_by_user = defaultdict(list)
        result = []
        # First, render all emails
        for submission in submissions:
            for speaker in submission.speakers.all():
                with suppress(
                    SendMailException
                ):  # This happens when there are template errors
                    mail = template.to_mail(
                        user=None,
                        event=self.event,
                        locale=submission.get_email_locale(speaker.locale),
                        context_kwargs={"submission": submission, "user": speaker},
                        commit=False,
                        allow_empty_address=True,
                    )
                    mails_by_user[speaker].append(mail)

        for user, user_mails in mails_by_user.items():
            # Second, deduplicate mails: we don't want speakers to receive the same
            # email twice, just because they have multiple submissions.
            mail_dict = {m.subject + m.text: m for m in user_mails}
            # Now we can create the emails and add the speakers to them
            for mail in mail_dict.values():
                mail.save()
                mail.to_users.add(user)
                result.append(mail)
        return result

    class Meta:
        model = MailTemplate
        fields = ["subject", "text", "bcc", "reply_to"]



import json

import pytest
from django_scopes import scope

from pretalx.api.serializers.room import RoomOrgaSerializer, RoomSerializer


@pytest.mark.django_db
def test_room_serializer(room):
    data = RoomSerializer(room).data
    assert set(data.keys()) == {
        "id",
        "guid",
        "name",
        "description",
        "capacity",
        "position",
        "url",
    }
    assert data["id"] == room.pk


@pytest.mark.django_db
def test_room_orga_serializer(room):
    with scope(event=room.event):
        data = RoomOrgaSerializer(room).data
        assert set(data.keys()) == {
            "id",
            "guid",
            "name",
            "description",
            "capacity",
            "position",
            "speaker_info",
            "availabilities",
            "url",
        }
        assert data["id"] == room.pk


@pytest.mark.django_db
def test_not_everybody_can_see_rooms(client, room):
    response = client.get(room.event.api_urls.rooms, follow=True)
    content = json.loads(response.content.decode())
    assert response.status_code == 200
    assert len(content["results"]) == 0, content


@pytest.mark.django_db
def test_everybody_can_see_published_rooms(client, room, slot):
    room.event.is_public = True
    room.event.save()
    response = client.get(room.event.api_urls.rooms, follow=True)
    content = json.loads(response.content.decode())
    assert response.status_code == 200
    assert len(content["results"]) == 1, content


@pytest.mark.django_db
def test_orga_can_see_room_speaker_info(orga_client, room):
    response = orga_client.get(room.event.api_urls.rooms, follow=True)
    content = json.loads(response.content.decode())
    assert response.status_code == 200
    assert len(content["results"]) == 1, content
    assert "speaker_info" in content["results"][0]



"""This command supersedes the Django-inbuilt runserver command.

It runs the local frontend server, if node is installed and the setting
is set.
"""
from pathlib import Path

from django.conf import settings
from django.core.management.commands.runserver import Command as Parent


class Command(Parent):
    def handle(self, *args, **options):
        if settings.VITE_DEV_MODE:
            # Start the vite server in the background
            from subprocess import Popen

            # run "npm start" in the frontend directory
            frontend_dir = (
                Path(__file__).parent.parent.parent.parent / "frontend/schedule-editor"
            )
            vite_server = Popen(["npm", "start"], cwd=frontend_dir)

        try:
            super().handle(*args, **options)
        finally:
            if settings.VITE_DEV_MODE:
                # Kill the vite server
                vite_server.kill()



import json

import pytest
from django_scopes import scope

from pretalx.api.serializers.review import ReviewSerializer


@pytest.mark.django_db
def test_review_serializer(review):
    with scope(event=review.event):
        data = ReviewSerializer(review).data
        assert set(data.keys()) == {
            "id",
            "answers",
            "submission",
            "user",
            "text",
            "score",
            "created",
            "updated",
        }
        assert data["submission"] == review.submission.code
        assert data["user"] == review.user.name
        assert data["answers"] == []


@pytest.mark.django_db
def test_anon_cannot_see_reviews(client, event, review):
    response = client.get(event.api_urls.reviews, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert len(content["results"]) == 0, content


@pytest.mark.django_db
def test_orga_can_see_reviews(orga_client, event, review):
    response = orga_client.get(event.api_urls.reviews, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert len(content["results"]) == 1


@pytest.mark.django_db
def test_orga_cannot_see_reviews_of_deleted_submission(orga_client, event, review):
    review.submission.state = "deleted"
    review.submission.save()
    response = orga_client.get(event.api_urls.reviews, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert len(content["results"]) == 0


@pytest.mark.django_db
def test_reviewer_can_see_reviews(review_client, event, review, other_review):
    response = review_client.get(event.api_urls.reviews, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert len(content["results"]) == 2, content


@pytest.mark.django_db
def test_reviewer_can_see_reviews_by_track(
    review_client, review_user, event, review, other_review, track, other_track
):
    review.submission.track = track
    review.submission.save()
    other_review.submission.track = other_track
    other_review.submission.save()
    review_user.teams.filter(is_reviewer=True).first().limit_tracks.add(track)

    response = review_client.get(event.api_urls.reviews, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert len(content["results"]) == 1, content


@pytest.mark.django_db
def test_reviewer_can_filter_by_submission(review_client, event, review, other_review):
    response = review_client.get(
        event.api_urls.reviews + f"?submission__code={review.submission.code}",
        follow=True,
    )
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert len(content["results"]) == 1, content


@pytest.mark.django_db
def test_reviewer_cannot_see_review_to_own_talk(
    review_user, review_client, event, review, other_review
):
    other_review.submission.speakers.add(review_user)
    response = review_client.get(event.api_urls.reviews, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert len(content["results"]) == 1, content



import pytest
from django.core.management import call_command
from django.db.models import Q
from django_scopes import scope

from pretalx.event.models import Event
from pretalx.schedule.models import Room, TalkSlot


@pytest.mark.django_db
def test_frab_import_minimal(administrator):
    assert Event.objects.count() == 0
    assert administrator.teams.count() == 0

    call_command("import_schedule", "tests/fixtures/frab_schedule_minimal.xml")

    assert Event.objects.count() == 1
    event = Event.objects.first()

    with scope(event=event):
        assert Room.objects.count() == 1
        assert Room.objects.all()[0].name == "Volkskundemuseum"

        assert TalkSlot.objects.count() == 2
        assert TalkSlot.objects.order_by("pk")[0].schedule.version == "1.99b 🍕"
        assert TalkSlot.objects.order_by("pk")[1].schedule.version is None
        assert event.name == "PrivacyWeek 2016"

        assert (
            administrator.teams.filter(
                Q(limit_events__in=[event]) | Q(all_events=True),
                can_change_event_settings=True,
            ).count()
            == 1
        )

        with pytest.raises(Exception):  # noqa
            call_command("import_schedule", "tests/fixtures/frab_schedule_minimal.xml")

        assert (
            administrator.teams.filter(
                Q(limit_events__in=[event]) | Q(all_events=True),
                can_change_event_settings=True,
            ).count()
            == 1
        )
        assert Event.objects.count() == 1
        assert TalkSlot.objects.count() == 2
        assert Room.objects.count() == 1
        event.organiser.teams.all().delete()

        call_command("import_schedule", "tests/fixtures/frab_schedule_minimal_2.xml")

        assert Room.objects.count() == 1
        assert Event.objects.count() == 1
        assert (
            administrator.teams.filter(
                Q(limit_events__in=[event]) | Q(all_events=True),
                can_change_event_settings=True,
            ).count()
            == 1
        )
        assert (
            TalkSlot.objects.count() == 5
        )  # 3 for the first talk, 2 for the second talk
        assert set(event.schedules.all().values_list("version", flat=True)) == {
            "1.99b 🍕",
            "1.99c 🍕",
            None,
        }


@pytest.mark.django_db
def test_frab_import_empty(administrator):
    assert Event.objects.count() == 0
    assert administrator.teams.count() == 0

    call_command("import_schedule", "tests/fixtures/frab_schedule_empty.xml")

    assert Event.objects.count() == 1
    event = Event.objects.first()

    with scope(event=event):
        assert Room.objects.count() == 1
        assert Room.objects.all()[0].name == "Volkskundemuseum"

        assert TalkSlot.objects.count() == 0

        assert (
            administrator.teams.filter(
                Q(limit_events__in=[event]) | Q(all_events=True),
                can_change_event_settings=True,
            ).count()
            == 1
        )



from pretalx.common.signals import EventPluginSignal

register_mail_placeholders = EventPluginSignal()
"""
This signal is sent out to get all known email text placeholders. Receivers should return
an instance of a subclass of pretalx.mail.placeholder.BaseMailTextPlaceholder or a list of these.

As with all event-plugin signals, the ``sender`` keyword argument will contain the event.
"""



import json

from django import forms
from django.utils.formats import get_format
from django.utils.translation import gettext as _
from django_scopes.forms import SafeModelChoiceField, SafeModelMultipleChoiceField

from pretalx.common.forms.fields import ImageField
from pretalx.common.forms.widgets import MarkdownWidget
from pretalx.common.mixins.forms import ReadOnlyFlag, RequestRequire
from pretalx.submission.models import Submission, SubmissionStates, SubmissionType


class SubmissionForm(ReadOnlyFlag, RequestRequire, forms.ModelForm):
    content_locale = forms.ChoiceField(label=_("Language"))

    def __init__(self, event, anonymise=False, **kwargs):
        self.event = event
        initial_slot = {}
        instance = kwargs.get("instance")
        if instance and instance.pk:
            slot = (
                instance.slots.filter(schedule__version__isnull=True)
                .select_related("room")
                .order_by("start")
                .first()
            )
            if slot:
                datetime_format = get_format("DATETIME_INPUT_FORMATS")[0]
                initial_slot = {
                    "room": slot.room,
                    "start": slot.local_start.strftime(datetime_format)
                    if slot.local_start
                    else "",
                    "end": slot.local_end.strftime(datetime_format)
                    if slot.real_end
                    else "",
                }
        if anonymise:
            kwargs.pop("initial", None)
            initial = {}
            instance = kwargs.pop("instance", None)
            previous_data = instance.anonymised
            for key in self._meta.fields:
                initial[key] = (
                    previous_data.get(key) or getattr(instance, key, None) or ""
                )
                if hasattr(initial[key], "all"):  # Tags, for the moment
                    initial[key] = initial[key].all()
            kwargs["initial"] = initial
        kwargs["initial"] = kwargs.get("initial") or {}
        kwargs["initial"].update(initial_slot)
        super().__init__(**kwargs)
        if "submission_type" in self.fields:
            self.fields["submission_type"].queryset = SubmissionType.objects.filter(
                event=event
            )
        if not self.event.tags.all().exists():
            self.fields.pop("tags", None)
        elif "tags" in self.fields:
            self.fields["tags"].queryset = self.event.tags.all()
            self.fields["tags"].required = False

        self.is_creating = False
        if not self.instance.pk:
            self.is_creating = True
            self.fields["speaker"] = forms.EmailField(
                label=_("Speaker email"),
                help_text=_(
                    "The email address of the speaker holding the session. They will be invited to create an account."
                ),
                required=False,
            )
            self.fields["speaker_name"] = forms.CharField(
                label=_("Speaker name"),
                help_text=_(
                    "The name of the speaker that should be displayed publicly."
                ),
                required=False,
            )
            if not anonymise:
                self.fields["state"] = forms.ChoiceField(
                    label=_("Proposal state"),
                    choices=SubmissionStates.get_choices(),
                    initial=SubmissionStates.SUBMITTED,
                )
        if not self.instance.pk or self.instance.state in (
            SubmissionStates.ACCEPTED,
            SubmissionStates.CONFIRMED,
        ):
            self.fields["room"] = forms.ModelChoiceField(
                required=False,
                queryset=event.rooms.all(),
                label=_("Room"),
                initial=initial_slot.get("room"),
            )
            self.fields["start"] = forms.DateTimeField(
                required=False,
                label=_("Start"),
                widget=forms.DateInput(
                    attrs={
                        "class": "datetimepickerfield",
                    }
                ),
                initial=initial_slot.get("start"),
            )
            self.fields["end"] = forms.DateTimeField(
                required=False,
                label=_("End"),
                widget=forms.DateInput(
                    attrs={
                        "class": "datetimepickerfield",
                    }
                ),
                initial=initial_slot.get("end"),
            )
        if "abstract" in self.fields:
            self.fields["abstract"].widget.attrs["rows"] = 2
        if not event.feature_flags["present_multiple_times"]:
            self.fields.pop("slot_count", None)
        if not event.feature_flags["use_tracks"]:
            self.fields.pop("track", None)
        elif "track" in self.fields:
            self.fields["track"].queryset = event.tracks.all()
        if "content_locale" in self.fields:
            if len(event.content_locales) == 1:
                self.fields.pop("content_locale")
            else:
                self.fields["content_locale"].choices = self.event.named_content_locales
        # If duration is not required, point out that the default is the session type's duration,
        # but only if there is more than one session type, because otherwise users will be
        # confused what that is.
        if (
            "duration" in self.fields
            and not self.fields["duration"].required
            and "submission_type" in self.fields
            and len(self.fields["submission_type"].queryset) > 1
        ):
            self.fields["duration"].help_text += " " + str(
                _("Leave empty to use the default duration for the session type.")
            )

    def clean(self):
        data = super().clean()
        start = data.get("start")
        end = data.get("end")
        if start and end and start > end:
            self.add_error(
                "end",
                forms.ValidationError(
                    _("The end time has to be after the start time."),
                ),
            )
        return data

    def save(self, *args, **kwargs):
        if "content_locale" not in self.fields:
            self.instance.content_locale = self.event.locale
        instance = super().save(*args, **kwargs)
        if self.is_creating:
            instance._set_state(self.cleaned_data["state"], force=True)
        else:
            if instance.pk and "duration" in self.changed_data:
                instance.update_duration()
            if instance.pk and "track" in self.changed_data:
                instance.update_review_scores()
            if "slot_count" in self.changed_data and "slot_count" in self.initial:
                instance.update_talk_slots()
        if (
            instance.state
            in (
                SubmissionStates.ACCEPTED,
                SubmissionStates.CONFIRMED,
            )
            and self.cleaned_data.get("room")
            and self.cleaned_data.get("start")
            and any(field in self.changed_data for field in ("room", "start", "end"))
        ):
            slot = (
                instance.slots.filter(schedule=instance.event.wip_schedule)
                .order_by("start")
                .first()
            )
            slot.room = self.cleaned_data.get("room")
            slot.start = self.cleaned_data.get("start")
            slot.end = self.cleaned_data.get("end")
            slot.save()
        return instance

    class Meta:
        model = Submission
        fields = [
            "title",
            "submission_type",
            "track",
            "tags",
            "abstract",
            "description",
            "notes",
            "internal_notes",
            "content_locale",
            "do_not_record",
            "duration",
            "slot_count",
            "image",
            "is_featured",
        ]
        widgets = {
            "tags": forms.SelectMultiple(attrs={"class": "select2"}),
            "track": forms.Select(attrs={"class": "select2"}),
            "submission_type": forms.Select(attrs={"class": "select2"}),
            "abstract": MarkdownWidget,
            "description": MarkdownWidget,
            "notes": MarkdownWidget,
        }
        field_classes = {
            "submission_type": SafeModelChoiceField,
            "tags": SafeModelMultipleChoiceField,
            "track": SafeModelChoiceField,
            "image": ImageField,
        }
        request_require = {
            "title",
            "abstract",
            "description",
            "notes",
            "image",
            "do_not_record",
            "content_locale",
        }


class AnonymiseForm(SubmissionForm):
    def __init__(self, *args, **kwargs):
        instance = kwargs.get("instance")
        if not instance or not instance.pk:
            raise Exception("Cannot anonymise unsaved submission.")
        kwargs["event"] = instance.event
        kwargs["anonymise"] = True
        super().__init__(*args, **kwargs)
        self._instance = instance
        to_be_removed = []
        for key, field in self.fields.items():
            try:
                field.plaintext = getattr(self._instance, key)
                field.required = False
            except AttributeError:
                to_be_removed.append(key)
        for key in to_be_removed:
            self.fields.pop(key)

    def save(self):
        anonymised_data = {"_anonymised": True}
        for key, value in self.cleaned_data.items():
            if value != getattr(self._instance, key, ""):
                anonymised_data[key] = value
        self._instance.anonymised_data = json.dumps(anonymised_data)
        self._instance.save(update_fields=["anonymised_data"])

    class Meta:
        model = Submission
        fields = [
            "title",
            "abstract",
            "description",
            "notes",
        ]
        request_require = fields


class SubmissionStateChangeForm(forms.Form):
    pending = forms.BooleanField(
        label=_("Mark the new state as 'pending'?"),
        help_text=_(
            "If you mark state changes as pending, they won't be visible to speakers right away. You can always apply pending changes for some or all proposals in one go once you're ready to make your decisions public."
        ),
        required=False,
        initial=False,
    )



import datetime as dt
import socket
from decimal import Decimal
from urllib.parse import urlparse

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.db.models import F, Q
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django_scopes.forms import SafeModelMultipleChoiceField
from i18nfield.fields import I18nFormField, I18nTextarea
from i18nfield.forms import I18nFormMixin, I18nModelForm

from pretalx.common.css import validate_css
from pretalx.common.forms.fields import ImageField
from pretalx.common.mixins.forms import (
    HierarkeyMixin,
    I18nHelpText,
    JsonSubfieldMixin,
    ReadOnlyFlag,
)
from pretalx.common.phrases import phrases
from pretalx.event.models.event import Event
from pretalx.orga.forms.widgets import HeaderSelect, MultipleLanguagesWidget
from pretalx.schedule.models import Availability, TalkSlot
from pretalx.submission.models import ReviewPhase, ReviewScore, ReviewScoreCategory

ENCRYPTED_PASSWORD_PLACEHOLDER = "*******"


class EventForm(ReadOnlyFlag, I18nHelpText, JsonSubfieldMixin, I18nModelForm):
    locales = forms.MultipleChoiceField(
        label=_("Active languages"),
        choices=[],
        widget=MultipleLanguagesWidget,
        help_text=_(
            "Users will be able to use pretalx in these languages, and you will be able to provide all texts in these"
            " languages. If you don't provide a text in the language a user selects, it will be shown in your event's"
            " default language instead."
        ),
    )
    content_locales = forms.MultipleChoiceField(
        label=_("Content languages"),
        choices=[],
        widget=forms.SelectMultiple(attrs={"class": "select2"}),
        help_text=_("Users will be able to submit proposals in these languages."),
    )
    custom_css_text = forms.CharField(
        required=False,
        widget=forms.Textarea(),
        label="",
        help_text=_("You can type in your CSS instead of uploading it, too."),
    )
    imprint_url = forms.URLField(
        label=_("Imprint URL"),
        help_text=_(
            "This should point e.g. to a part of your website that has your contact details and legal information."
        ),
        required=False,
    )
    show_schedule = forms.BooleanField(
        label=_("Show schedule publicly"),
        help_text=_(
            "Unset to hide your schedule, e.g. if you want to use the HTML export exclusively."
        ),
        required=False,
    )
    schedule = forms.ChoiceField(
        label=_("Schedule display format"),  # TODO show small preview / icons
        choices=(
            ("grid", _("Grid")),
            ("list", _("List")),
        ),
        required=True,
    )
    show_featured = forms.ChoiceField(
        label=_("Show featured sessions"),
        choices=(
            ("never", _("Never")),
            ("pre_schedule", _("Until the first schedule is released")),
            ("always", _("Always")),
        ),
        help_text=_(
            "Marking sessions as 'featured' is a good way to show them before the first schedule release, or to highlight them once the schedule is visible."
        ),
        required=True,
    )
    use_feedback = forms.BooleanField(
        label=_("Enable anonymous feedback"),
        help_text=_(
            "Attendees will be able to send in feedback after a session is over."
        ),
        required=False,
    )
    export_html_on_release = forms.BooleanField(
        label=_("Generate HTML export on schedule release"),
        help_text=_(
            "The static HTML export will be provided as a .zip archive on the schedule export page."
        ),
        required=False,
    )
    html_export_url = forms.URLField(
        label=_("HTML Export URL"),
        help_text=_(
            "If you publish your schedule via the HTML export, you will want the correct absolute URL to be set in various places. "
            "Please only set this value once you have published your schedule. Should end with a slash."
        ),
        required=False,
    )
    header_pattern = forms.ChoiceField(
        label=_("Frontpage header pattern"),
        help_text=_(
            "Choose how the frontpage header banner will be styled if you don't upload an image. Pattern source: "
            '<a href="http://www.heropatterns.com/">heropatterns.com</a>, CC BY 4.0.'
        ),
        choices=(
            ("", _("Plain")),
            ("pcb", _("Circuits")),
            ("bubbles", _("Circles")),
            ("signal", _("Signal")),
            ("topo", _("Topography")),
            ("graph", _("Graph Paper")),
        ),
        required=False,
        widget=HeaderSelect,
    )
    meta_noindex = forms.BooleanField(
        label=_("Ask search engines not to index the event pages"), required=False
    )

    def __init__(self, *args, **kwargs):
        self.is_administrator = kwargs.pop("is_administrator", False)
        super().__init__(*args, **kwargs)
        self.initial["locales"] = self.instance.locale_array.split(",")
        self.initial["content_locales"] = self.instance.content_locale_array.split(",")
        year = str(now().year)
        self.fields["show_featured"].help_text = (
            str(self.fields["show_featured"].help_text)
            + " "
            + str(_("You can find the page <a {href}>here</a>.")).format(
                href=f'href="{self.instance.urls.featured}"'
            )
        )
        self.fields["name"].widget.attrs["placeholder"] = (
            _("The name of your conference, e.g. My Conference") + " " + year
        )
        if not self.is_administrator:
            self.fields["slug"].disabled = True
            self.fields["slug"].help_text = _(
                "Please contact your administrator if you need to change the short name of your event."
            )
        self.fields["primary_color"].widget.attrs["placeholder"] = _(
            "A color hex value, e.g. #ab01de"
        )
        self.fields["primary_color"].widget.attrs["class"] = "colorpickerfield"
        self.fields["date_to"].help_text = _(
            "Any sessions you have scheduled already will be moved if you change the event dates. You will have to release a new schedule version to notify all speakers."
        )
        self.fields["locales"].choices = [
            choice
            for choice in settings.LANGUAGES
            if settings.LANGUAGES_INFORMATION[choice[0]].get("visible", True)
            or choice[0] in self.instance.plugin_locales
        ]
        self.fields["content_locales"].choices = self.instance.available_content_locales

    def clean_custom_domain(self):
        data = self.cleaned_data["custom_domain"]
        if not data:
            return data
        data = data.lower()
        if data in [urlparse(settings.SITE_URL).hostname, settings.SITE_URL]:
            raise ValidationError(
                _("Please do not choose the default domain as custom event domain.")
            )
        if not data.startswith("https://"):
            data = data[len("http://") :] if data.startswith("http://") else data
            data = "https://" + data
        data = data.rstrip("/")
        try:
            socket.gethostbyname(data[len("https://") :])
        except OSError:
            raise forms.ValidationError(
                _(
                    'The domain "{domain}" does not have a name server entry at this time. Please make sure the domain is working before configuring it here.'
                ).format(domain=data)
            )
        return data

    def clean_custom_css(self):
        if self.cleaned_data.get("custom_css") or self.files.get("custom_css"):
            css = self.cleaned_data["custom_css"] or self.files["custom_css"]
            if self.is_administrator:
                return css
            try:
                validate_css(css.read())
                return css
            except IsADirectoryError:
                self.instance.custom_css = None
                self.instance.save(update_fields=["custom_css"])
        else:
            self.instance.custom_css = None
            self.instance.save(update_fields=["custom_css"])
        return None

    def clean_custom_css_text(self):
        css = self.cleaned_data.get("custom_css_text").strip()
        if not css or self.is_administrator:
            return css
        validate_css(css)
        return css

    def clean(self):
        data = super().clean()
        date_from = data.get("date_from")
        date_to = data.get("date_to")
        if date_from and date_to and date_from > date_to:
            error = forms.ValidationError(
                _("The event end cannot be before the start.")
            )
            self.add_error("date_from", error)
        if data.get("locale") not in data.get("locales", []):
            error = forms.ValidationError(
                _("Your default language needs to be one of your active languages."),
            )
            self.add_error("locale", error)
        return data

    def save(self, *args, **kwargs):
        self.instance.locale_array = ",".join(self.cleaned_data["locales"])
        self.instance.content_locale_array = ",".join(
            self.cleaned_data["content_locales"]
        )
        if any(key in self.changed_data for key in ("date_from", "date_to")):
            self.change_dates()
        if "timezone" in self.changed_data:
            self.change_timezone()
        result = super().save(*args, **kwargs)
        css_text = self.cleaned_data["custom_css_text"]
        if css_text:
            self.instance.custom_css.save(
                self.instance.slug + ".css", ContentFile(css_text)
            )
        return result

    def change_dates(self):
        """Changes dates of current WIP slots, or deschedules them."""

        old_instance = Event.objects.get(pk=self.instance.pk)
        if not self.instance.wip_schedule.talks.filter(start__isnull=False).exists():
            return
        new_date_from = self.cleaned_data["date_from"]
        new_date_to = self.cleaned_data["date_to"]
        start_delta = new_date_from - old_instance.date_from
        end_delta = new_date_to - old_instance.date_to
        shortened = (new_date_to - new_date_from) < (
            old_instance.date_to - old_instance.date_from
        )

        if start_delta and end_delta:
            # The event was moved, and we will move all talks with it.
            self._move_by(start_delta)

        # Otherwise, the event got longer, no need to do anything.
        # We *could* move all talks towards the new start date, but I'm
        # not convinced that this is the actual use case.
        # I think it's more likely that people add a new day to the start.
        if shortened:
            # The event was shortened, de-schedule all talks outside the range
            self.instance.wip_schedule.talks.filter(
                Q(start__date__gt=new_date_to) | Q(start__date__lt=new_date_from),
            ).update(start=None, end=None, room=None)
            Availability.objects.filter(
                Q(end__date__gt=new_date_to) | Q(start__date__lt=new_date_from),
                event=self.instance.event,
            ).delete()

    def change_timezone(self):
        """Changes times of all current wip slots, on the assumption that a
        change in timezone is usually not intentional, and people would like to
        keep the apparent time rather the absolute one."""

        old_instance = Event.objects.get(pk=self.instance.pk)
        first_slot = self.instance.wip_schedule.talks.filter(
            start__isnull=False
        ).first()
        if not first_slot:
            return

        def make_naive(moment):
            return dt.datetime(
                year=moment.year,
                month=moment.month,
                day=moment.day,
                hour=moment.hour,
                minute=moment.minute,
            )

        old_start = make_naive(first_slot.start.astimezone(old_instance.tz))
        new_start = make_naive(first_slot.start.astimezone(self.instance.tz))

        delta = old_start - new_start
        if delta:
            self._move_by(delta, past=True)

    def _move_by(self, delta, past=False):
        if past:
            talk_queryset = TalkSlot.objects.filter(schedule__event=self.instance)
        else:
            talk_queryset = self.instance.wip_schedule.talks
        for key in ("start", "end"):
            filt = {f"{key}__isnull": False}
            update = {key: F(key) + delta}
            talk_queryset.filter(**filt).update(**update)
            Availability.objects.filter(event=self.instance).filter(**filt).update(
                **update
            )

    class Meta:
        model = Event
        fields = [
            "name",
            "slug",
            "date_from",
            "date_to",
            "timezone",
            "email",
            "locale",
            "custom_domain",
            "primary_color",
            "custom_css",
            "logo",
            "header_image",
            "landing_page_text",
            "featured_sessions_text",
        ]
        field_classes = {
            "logo": ImageField,
            "header_image": ImageField,
        }
        widgets = {
            "date_from": forms.DateInput(attrs={"class": "datepickerfield"}),
            "date_to": forms.DateInput(
                attrs={"class": "datepickerfield", "data-date-after": "#id_date_from"}
            ),
            "locale": forms.Select(attrs={"class": "select2"}),
            "timezone": forms.Select(attrs={"class": "select2"}),
        }
        json_fields = {
            "imprint_url": "display_settings",
            "show_schedule": "feature_flags",
            "schedule": "display_settings",
            "show_featured": "feature_flags",
            "use_feedback": "feature_flags",
            "export_html_on_release": "feature_flags",
            "html_export_url": "display_settings",
            "header_pattern": "display_settings",
            "meta_noindex": "display_settings",
        }


class MailSettingsForm(
    ReadOnlyFlag, I18nFormMixin, I18nHelpText, JsonSubfieldMixin, forms.Form
):
    reply_to = forms.EmailField(
        label=_("Contact address"),
        help_text=_(
            "Reply-To address. If this setting is empty and you have no custom sender, your event email address will be used as Reply-To address."
        ),
        required=False,
    )
    subject_prefix = forms.CharField(
        label=_("Mail subject prefix"),
        help_text=_(
            "The prefix will be prepended to outgoing mail subjects in [brackets]."
        ),
        required=False,
    )
    signature = forms.CharField(
        label=_("Mail signature"),
        help_text=_(
            'The signature will be added to outgoing mails, preceded by "-- ".'
        ),
        required=False,
        widget=forms.Textarea,
    )
    smtp_use_custom = forms.BooleanField(
        label=_("Use custom SMTP server"),
        help_text=_(
            "All mail related to your event will be sent over the SMTP server specified by you."
        ),
        required=False,
    )
    mail_from = forms.EmailField(
        label=_("Sender address"),
        help_text=_("Sender address for outgoing emails."),
        required=False,
    )
    smtp_host = forms.CharField(label=_("Hostname"), required=False)
    smtp_port = forms.IntegerField(label=_("Port"), required=False)
    smtp_username = forms.CharField(label=_("Username"), required=False)
    smtp_password = forms.CharField(
        label=_("Password"),
        required=False,
        widget=forms.PasswordInput(
            attrs={
                "autocomplete": "new-password"  # see https://bugs.chromium.org/p/chromium/issues/detail?id=370363#c7
            }
        ),
    )
    smtp_use_tls = forms.BooleanField(
        label=_("Use STARTTLS"),
        help_text=_("Commonly enabled on port 587."),
        required=False,
    )
    smtp_use_ssl = forms.BooleanField(
        label=_("Use SSL"), help_text=_("Commonly enabled on port 465."), required=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_encrypted_password_placeholder()

    def set_encrypted_password_placeholder(self):
        if self.initial.get("smtp_password"):
            self.fields["smtp_password"].widget.attrs[
                "placeholder"
            ] = ENCRYPTED_PASSWORD_PLACEHOLDER

    def clean(self):
        data = self.cleaned_data
        if not data.get("smtp_use_custom"):
            # We don't need to validate all the rest when we don't use a custom email server
            return data

        if not data.get("smtp_password") and data.get("smtp_username"):
            # Leave password unchanged if the username is set and the password field is empty.
            # This makes it impossible to set an empty password as long as a username is set, but
            # Python's smtplib does not support password-less schemes anyway.
            data["smtp_password"] = self.initial.get("smtp_password")

        if not data.get("mail_from"):
            self.add_error(
                "mail_from",
                ValidationError(
                    _(
                        "You have to provide a sender address if you use a custom SMTP server."
                    )
                ),
            )
        if data.get("smtp_use_tls") and data.get("smtp_use_ssl"):
            self.add_error(
                "smtp_use_tls",
                ValidationError(
                    _(
                        "You can activate either SSL or STARTTLS security, but not both at the same time."
                    )
                ),
            )
        uses_encryption = data.get("smtp_use_tls") or data.get("smtp_use_ssl")
        localhost_names = [
            "127.0.0.1",
            "::1",
            "[::1]",
            "localhost",
            "localhost.localdomain",
        ]
        if not uses_encryption and not data.get("smtp_host") in localhost_names:
            self.add_error(
                "smtp_host",
                ValidationError(
                    _(
                        "You have to activate either SSL or STARTTLS security if you use a non-local mailserver due to data protection reasons. "
                        "Your administrator can add an instance-wide bypass. If you use this bypass, please also adjust your Privacy Policy."
                    )
                ),
            )

    class Meta:
        json_fields = {
            "reply_to": "mail_settings",
            "subject_prefix": "mail_settings",
            "signature": "mail_settings",
            "smtp_use_custom": "mail_settings",
            "mail_from": "mail_settings",
            "smtp_host": "mail_settings",
            "smtp_port": "mail_settings",
            "smtp_username": "mail_settings",
            "smtp_password": "mail_settings",
            "smtp_use_tls": "mail_settings",
            "smtp_use_ssl": "mail_settings",
        }


class ReviewSettingsForm(
    ReadOnlyFlag,
    I18nFormMixin,
    I18nHelpText,
    HierarkeyMixin,
    JsonSubfieldMixin,
    forms.Form,
):
    score_mandatory = forms.BooleanField(
        label=_("Require a review score"), required=False
    )
    text_mandatory = forms.BooleanField(
        label=_("Require a review text"), required=False
    )
    score_format = forms.ChoiceField(
        label=_("Score display"),
        required=True,
        choices=(
            ("words_numbers", _("Text and score, e.g “Good (3)”")),
            ("numbers_words", _("Score and text, e.g “3 (good)”")),
            ("numbers", _("Just scores")),
            ("words", _("Just text")),
        ),
        initial="words_numbers",
        help_text=_(
            "This is how the score will look like on the review interface. The dashboard will always show numerical scores."
        ),
        widget=forms.RadioSelect,
    )
    aggregate_method = forms.ChoiceField(
        label=_("Score aggregation method"),
        required=True,
        choices=(("median", _("Median")), ("mean", _("Average (mean)"))),
        widget=forms.RadioSelect,
    )
    review_help_text = I18nFormField(
        label=_("Help text for reviewers"),
        help_text=_(
            "This text will be shown at the top of every review, as long as reviews can be created or edited."
        )
        + " "
        + phrases.base.use_markdown,
        widget=I18nTextarea,
        required=False,
    )

    class Meta:
        json_fields = {
            "score_mandatory": "review_settings",
            "text_mandatory": "review_settings",
            "aggregate_method": "review_settings",
            "score_format": "review_settings",
        }
        hierarkey_fields = ("review_help_text",)


class WidgetSettingsForm(JsonSubfieldMixin, forms.Form):
    show_widget_if_not_public = forms.BooleanField(
        label=_("Show the widget even if the schedule is not public"),
        help_text=_(
            "Set to allow external pages to show the schedule widget, even if the schedule is not shown here using pretalx."
        ),
        required=False,
    )

    class Meta:
        json_fields = {"show_widget_if_not_public": "feature_flags"}


class WidgetGenerationForm(forms.ModelForm):
    schedule_display = forms.ChoiceField(
        label=_("Schedule display format"),
        choices=(
            ("grid", _("Grid")),
            ("list", _("List")),
        ),
        required=True,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["locale"].label = _("Widget language")

    class Meta:
        model = Event
        fields = ["locale"]
        widgets = {"locale": forms.Select(attrs={"class": "select2"})}


class ReviewPhaseForm(I18nHelpText, I18nModelForm):
    def __init__(self, *args, event=None, **kwargs):
        super().__init__(*args, **kwargs)

    def clean(self):
        data = super().clean()
        if data.get("start") and data.get("end") and data["start"] > data["end"]:
            self.add_error(
                "end",
                forms.ValidationError(
                    _("The end of a phase has to be after its start.")
                ),
            )
        return data

    class Meta:
        model = ReviewPhase
        fields = [
            "name",
            "start",
            "end",
            "can_review",
            "proposal_visibility",
            "can_see_speaker_names",
            "can_see_reviewer_names",
            "can_change_submission_state",
            "can_see_other_reviews",
            "can_tag_submissions",
            "speakers_can_change_submissions",
        ]
        widgets = {
            "start": forms.DateInput(attrs={"class": "datetimepickerfield"}),
            "end": forms.DateInput(attrs={"class": "datetimepickerfield"}),
        }


def strip_zeroes(value):
    if not isinstance(value, Decimal):
        return value
    value = str(value)
    return Decimal(value.rstrip("0"))


class ReviewScoreCategoryForm(I18nHelpText, I18nModelForm):
    new_scores = forms.CharField(required=False, initial="")

    def __init__(self, *args, event=None, **kwargs):
        self.event = event
        super().__init__(*args, **kwargs)
        if not event or not event.feature_flags["use_tracks"]:
            self.fields.pop("limit_tracks")
        else:
            self.fields["limit_tracks"].queryset = event.tracks.all()
        ids = self.data.get(self.prefix + "-new_scores")
        self.new_label_ids = ids.strip(",").split(",") if ids else []
        for label_id in self.new_label_ids:
            self._add_score_fields(label_id=label_id)

        self.label_fields = []
        if self.instance.id:
            scores = self.instance.scores.all()
            for score in scores:
                self.label_fields.append(
                    {
                        "score": score,
                        "label_field": score._meta.get_field("label").formfield(
                            initial=score.label
                        ),
                        "value_field": score._meta.get_field("value").formfield(
                            initial=strip_zeroes(score.value), required=False
                        ),
                    }
                )
        for score in self.label_fields:
            score_id = score["score"].id
            self.fields[f"value_{score_id}"] = score["value_field"]
            self.fields[f"label_{score_id}"] = score["label_field"]

    def _add_score_fields(self, label_id):
        self.fields[f"value_{label_id}"] = ReviewScore._meta.get_field(
            "value"
        ).formfield()
        self.fields[f"label_{label_id}"] = ReviewScore._meta.get_field(
            "label"
        ).formfield()

    def get_label_fields(self):
        for score in self.label_fields:
            score_id = score["score"].id
            yield (self[f"value_{score_id}"], self[f"label_{score_id}"])

    def save(self, *args, **kwargs):
        instance = super().save(*args, **kwargs)
        for score in self.label_fields:
            score_id = score["score"].id
            if any(f"_{score_id}" in changed for changed in self.changed_data):
                value = self.cleaned_data.get(f"value_{score_id}")
                label = self.cleaned_data.get(f"label_{score_id}")
                if value is None or value == "":
                    score["score"].delete()
                else:
                    score["score"].value = value
                    score["score"].label = label
                    score["score"].save()
        for score in self.new_label_ids:
            value = self.cleaned_data.get(f"value_{score}")
            label = self.cleaned_data.get(f"label_{score}")
            if (value is not None) and label:
                ReviewScore.objects.create(
                    category=self.instance, value=value, label=label
                )
        return instance

    class Meta:
        model = ReviewScoreCategory
        fields = (
            "name",
            "is_independent",
            "weight",
            "required",
            "active",
            "limit_tracks",
        )
        field_classes = {
            "limit_tracks": SafeModelMultipleChoiceField,
        }
        widgets = {"limit_tracks": forms.SelectMultiple(attrs={"class": "select2"})}



import pytest
from django_scopes import scope


@pytest.mark.parametrize("url", ("v1.en.js", "v1.json", "v1.css", "v2.json"))
@pytest.mark.parametrize(
    "show_schedule,show_widget_if_not_public,expected",
    (
        (True, False, 200),
        (True, True, 200),
        (False, False, 404),
        (False, True, 200),
    ),
)
@pytest.mark.django_db
def test_widget_pages(
    event,
    schedule,
    client,
    url,
    show_schedule,
    show_widget_if_not_public,
    expected,
    slot,
    other_slot,
):
    event.feature_flags["show_schedule"] = show_schedule
    event.feature_flags["show_widget_if_not_public"] = show_widget_if_not_public
    event.save()
    response = client.get(event.urls.schedule + "widget/" + url, follow=True)
    assert response.status_code == expected


@pytest.mark.django_db
@pytest.mark.parametrize(
    "version,queries",
    (
        ("1", 8),
        ("2", 11),
    ),
)
def test_widget_data(
    client,
    event,
    schedule,
    slot,
    other_slot,
    version,
    queries,
    django_assert_num_queries,
):
    event.feature_flags["show_schedule"] = True
    event.save()
    with django_assert_num_queries(queries):
        response = client.get(
            event.urls.schedule + f"widget/v{version}.json", follow=True
        )
    assert response.status_code == 200


@pytest.mark.django_db
def test_versioned_widget_data(client, event, schedule, slot):
    with scope(event=event):
        event.wip_schedule.freeze("new")

    response = client.get(event.urls.schedule + f"widget/v2.json?v={schedule.version}")
    assert response.status_code == 200


@pytest.mark.django_db
def test_bogus_versioned_widget_data(client, event, schedule, slot):
    response = client.get(event.urls.schedule + "widget/v2.json?v=nopedinope")
    assert response.status_code == 200


@pytest.mark.django_db
def test_anon_cannot_access_wip_schedule(client, event, schedule, slot):
    response = client.get(event.urls.schedule + "widget/v2.json?v=wip")
    assert response.status_code == 404


@pytest.mark.django_db
def test_orga_can_access_wip_schedule(orga_client, event, schedule, slot):
    response = orga_client.get(event.urls.schedule + "widget/v2.json?v=wip")
    assert response.status_code == 200


@pytest.mark.parametrize(
    "locale,expected",
    (
        ("lo", 404),
        ("en", 200),
    ),
)
@pytest.mark.django_db
def test_widget_wrong_locale(event, schedule, client, locale, expected):
    response = client.get(event.urls.schedule + "widget/v1." + locale + ".js")
    assert response.status_code == expected


@pytest.mark.django_db
def test_widget_with_primary_color(event, schedule, client):
    event.primary_color = "#abcdef"
    event.save()
    response = client.get(event.urls.schedule + "widget/v1.css")
    assert event.primary_color in response.content.decode()



import pytest
from django_scopes import scope

from pretalx.schedule.utils import guess_schedule_version


@pytest.mark.django_db
@pytest.mark.parametrize(
    "previous,suggestion",
    (
        (None, "0.1"),
        ("0.1", "0.2"),
        ("0,2", "0,3"),
        ("0-3", "0-4"),
        ("0_4", "0_5"),
        ("1.0.1", "1.0.2"),
        ("Nichtnumerisch", ""),
        ("1.someting", ""),
        ("something.1", "something.2"),
    ),
)
def test_schedule_version_guessing(event, previous, suggestion):
    with scope(event=event):
        if previous:
            event.release_schedule(previous)
        assert guess_schedule_version(event) == suggestion



import pytest
from django.core.management import call_command


@pytest.mark.django_db
def test_no_outstanding_migrations():
    result = call_command(
        "makemigrations", "--dry-run", "--check", "--no-input", "-v", "0"
    )
    assert result is None



import json

import pytest
from django_scopes import scope

from pretalx.api.serializers.submission import (
    SubmissionOrgaSerializer,
    SubmissionSerializer,
    TagSerializer,
)


@pytest.mark.django_db
def test_submission_slot_serializer(slot):
    with scope(event=slot.submission.event):
        data = SubmissionSerializer(
            slot.submission, context={"event": slot.submission.event}
        ).data
        assert set(data.keys()) == {
            "code",
            "speakers",
            "title",
            "submission_type",
            "submission_type_id",
            "state",
            "abstract",
            "description",
            "duration",
            "slot_count",
            "do_not_record",
            "is_featured",
            "content_locale",
            "slot",
            "image",
            "track",
            "track_id",
            "resources",
        }
        assert set(data["slot"].keys()) == {"start", "end", "room", "room_id"}
        assert data["slot"]["room"] == slot.room.name


@pytest.mark.django_db
def test_tag_serializer(tag):
    with scope(event=tag.event):
        data = TagSerializer(tag, context={"event": tag.event}).data
        assert set(data.keys()) == {
            "id",
            "tag",
            "description",
            "color",
        }


@pytest.mark.django_db
def test_submission_serializer_for_organiser(submission, orga_user, resource, tag):
    with scope(event=submission.event):
        submission.tags.add(tag)
        data = SubmissionOrgaSerializer(
            submission,
            event=submission.event,
            can_view_speakers=True,
        ).data
        assert set(data.keys()) == {
            "code",
            "speakers",
            "title",
            "submission_type",
            "submission_type_id",
            "state",
            "pending_state",
            "abstract",
            "description",
            "duration",
            "slot_count",
            "do_not_record",
            "is_featured",
            "content_locale",
            "slot",
            "image",
            "answers",
            "track",
            "track_id",
            "notes",
            "internal_notes",
            "created",
            "resources",
            "tags",
            "tag_ids",
        }
        assert isinstance(data["speakers"], list)
        assert data["speakers"][0] == {
            "name": submission.speakers.first().name,
            "code": submission.speakers.first().code,
            "email": submission.speakers.first().email,
            "biography": submission.speakers.first()
            .event_profile(submission.event)
            .biography,
            "avatar": None,
        }
        assert data["tags"] == [tag.tag]
        assert data["tag_ids"] == [tag.id]
        assert data["submission_type"] == str(submission.submission_type.name)
        assert data["slot"] is None
        assert (
            data["created"]
            == submission.created.astimezone(submission.event.tz).isoformat()
        )
        assert data["resources"] == [
            {
                "resource": "http://localhost" + resource.resource.url,
                "description": resource.description,
            }
        ]


@pytest.mark.django_db
def test_submission_serializer(submission, resource):
    with scope(event=submission.event):
        data = SubmissionSerializer(
            submission, context={"event": submission.event}
        ).data
        assert set(data.keys()) == {
            "code",
            "speakers",
            "title",
            "submission_type",
            "submission_type_id",
            "state",
            "abstract",
            "description",
            "duration",
            "slot_count",
            "do_not_record",
            "is_featured",
            "content_locale",
            "slot",
            "image",
            "track",
            "track_id",
            "resources",
        }
        assert isinstance(data["speakers"], list)
        assert data["speakers"] == []
        assert data["submission_type"] == str(submission.submission_type.name)
        assert data["slot"] is None
        assert data["resources"] == [
            {
                "resource": "http://localhost" + resource.resource.url,
                "description": resource.description,
            }
        ]


@pytest.mark.django_db
def test_can_only_see_public_submissions(
    client, slot, accepted_submission, rejected_submission, submission
):
    response = client.get(submission.event.api_urls.submissions, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 1
    assert content["results"][0]["title"] == slot.submission.title


@pytest.mark.django_db
def test_can_only_see_public_submissions_if_public_schedule(
    client, slot, accepted_submission, rejected_submission, submission, answer
):
    submission.event.feature_flags["show_schedule"] = False
    submission.event.save()
    response = client.get(submission.event.api_urls.submissions, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 0
    assert all(submission["answers"] == [] for submission in content["results"])


@pytest.mark.django_db
def test_orga_can_see_all_submissions(
    orga_client, slot, accepted_submission, rejected_submission, submission, answer
):
    response = orga_client.get(
        submission.event.api_urls.submissions + "?questions=all", follow=True
    )
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 4
    assert content["results"][0]["title"] == slot.submission.title
    assert len(
        [submission for submission in content["results"] if submission["answers"] == []]
    )
    assert len(
        [submission for submission in content["results"] if submission["answers"] != []]
    )


@pytest.mark.django_db
def test_orga_can_see_all_submissions_wrong_question(
    orga_client, slot, accepted_submission, rejected_submission, submission, answer
):
    response = orga_client.get(
        submission.event.api_urls.submissions + "?questions=1212112", follow=True
    )
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 4
    assert content["results"][0]["title"] == slot.submission.title
    assert not len(
        [submission for submission in content["results"] if submission["answers"] != []]
    )


@pytest.mark.django_db
@pytest.mark.parametrize(
    "is_visible_to_reviewers,is_reviewer,length",
    (
        (True, True, 1),
        (False, True, 0),
        (True, False, 1),
        (False, False, 1),
    ),
)
def test_answer_is_visible_to_reviewers(
    orga_client,
    review_user,
    submission,
    answer,
    event,
    is_visible_to_reviewers,
    is_reviewer,
    length,
):
    if is_reviewer:
        orga_client.force_login(review_user)

    with scope(event=event):
        question = answer.question
        question.is_visible_to_reviewers = is_visible_to_reviewers
        question.save()

    response = orga_client.get(
        submission.event.api_urls.submissions + f"?questions={answer.question_id}",
        follow=True,
    )
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 1
    assert content["results"][0]["title"] == submission.title
    assert len(content["results"][0]["answers"]) == length


@pytest.mark.django_db
def test_orga_can_see_all_submissions_even_nonpublic(
    orga_client, slot, accepted_submission, rejected_submission, submission
):
    submission.event.feature_flags["show_schedule"] = False
    submission.event.save()
    response = orga_client.get(submission.event.api_urls.submissions, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 4
    assert content["results"][0]["title"] == slot.submission.title


@pytest.mark.django_db
def test_only_see_talks_when_a_release_exists(
    orga_client, confirmed_submission, rejected_submission, submission
):
    response = orga_client.get(submission.event.api_urls.talks, follow=True)
    content = json.loads(response.content.decode())
    assert response.status_code == 200
    assert content["count"] == 0


@pytest.mark.django_db
def test_can_only_see_public_talks(
    event, client, slot, accepted_submission, rejected_submission, submission
):
    response = client.get(submission.event.api_urls.talks, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 1
    assert content["results"][0]["title"] == slot.submission.title
    with scope(event=event):
        speaker = slot.submission.speakers.first()
        assert content["results"][0]["speakers"][0]["name"] == speaker.name
        assert (
            content["results"][0]["speakers"][0]["biography"]
            == speaker.event_profile(event).biography
        )


@pytest.mark.django_db
def test_can_only_see_public_talks_if_public_schedule(
    client, slot, accepted_submission, rejected_submission, submission
):
    submission.event.feature_flags["show_schedule"] = False
    submission.event.save()
    response = client.get(submission.event.api_urls.talks, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 0


@pytest.mark.django_db
def test_orga_can_see_all_talks(
    orga_client, slot, accepted_submission, rejected_submission, submission
):
    response = orga_client.get(submission.event.api_urls.talks, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 1
    assert content["results"][0]["title"] == slot.submission.title


@pytest.mark.django_db
def test_orga_can_see_all_talks_even_nonpublic(
    orga_client, slot, accepted_submission, rejected_submission, submission
):
    submission.event.feature_flags["show_schedule"] = False
    submission.event.save()
    response = orga_client.get(submission.event.api_urls.talks, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 1
    assert content["results"][0]["title"] == slot.submission.title


@pytest.mark.django_db
def test_reviewer_cannot_see_speakers_and_anonymised_content(
    orga_client,
    review_user,
    submission,
    event,
):
    with scope(event=event):
        submission.event.active_review_phase.can_see_speaker_names = False
        submission.event.active_review_phase.save()
        submission.anonymised_data = json.dumps({"description": "CENSORED!"})
        submission.save()
    response = orga_client.get(submission.event.api_urls.submissions, follow=True)
    assert response.status_code == 200
    content = json.loads(response.content.decode())
    content = content["results"][0]
    assert len(content["speakers"]) == 1
    assert content["description"] != "CENSORED!"
    assert content["abstract"] == submission.abstract

    orga_client.force_login(review_user)

    response = orga_client.get(submission.event.api_urls.submissions, follow=True)
    assert response.status_code == 200
    content = json.loads(response.content.decode())
    content = content["results"][0]
    assert content["speakers"] == []
    assert content["description"] == "CENSORED!"
    assert content["abstract"] == submission.abstract


@pytest.mark.django_db
def test_cannot_see_tags(client, tag):
    response = client.get(tag.event.api_urls.tags, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 0


@pytest.mark.django_db
def test_orga_can_see_tags(orga_client, tag):
    response = orga_client.get(tag.event.api_urls.tags, follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["count"] == 1
    assert content["results"][0]["tag"] == tag.tag


@pytest.mark.django_db
def test_orga_can_see_single_tag(orga_client, tag):
    response = orga_client.get(tag.event.api_urls.tags + f"{tag.tag}/", follow=True)
    content = json.loads(response.content.decode())

    assert response.status_code == 200
    assert content["tag"] == tag.tag



import pytest

from pretalx.person.models.information import resource_path


@pytest.mark.django_db
def test_information_resource_path(information):
    assert resource_path(information, "foo").startswith(
        f"{information.event.slug}/speaker_information/foo"
    )



import pytest
from django.conf import settings
from django.core import mail as djmail
from django.core.files.uploadedfile import SimpleUploadedFile
from django.urls import reverse
from django_scopes import scope
from rest_framework.authtoken.models import Token

from pretalx.submission.models import SubmissionStates


@pytest.mark.django_db
def test_can_see_submission_list(speaker_client, submission):
    response = speaker_client.get(submission.event.urls.user_submissions, follow=True)
    assert response.status_code == 200
    assert submission.title in response.content.decode()


@pytest.mark.django_db
def test_can_see_submission(speaker_client, submission):
    response = speaker_client.get(submission.urls.user_base, follow=True)
    assert response.status_code == 200
    assert submission.title in response.content.decode()


@pytest.mark.django_db
def test_cannot_see_other_submission(speaker_client, other_submission):
    response = speaker_client.get(other_submission.urls.user_base, follow=True)
    assert response.status_code == 404


@pytest.mark.django_db
def test_can_confirm_submission(speaker_client, accepted_submission):
    response = speaker_client.get(accepted_submission.urls.confirm, follow=True)
    accepted_submission.refresh_from_db()
    assert response.status_code == 200
    assert accepted_submission.state == SubmissionStates.ACCEPTED
    response = speaker_client.post(accepted_submission.urls.confirm, follow=True)
    accepted_submission.refresh_from_db()
    assert response.status_code == 200
    assert accepted_submission.state == SubmissionStates.CONFIRMED


@pytest.mark.django_db
def test_cannot_confirm_confirmed_submission(speaker_client, confirmed_submission):
    response = speaker_client.get(confirmed_submission.urls.confirm, follow=True)
    confirmed_submission.refresh_from_db()
    assert response.status_code == 200
    assert confirmed_submission.state == SubmissionStates.CONFIRMED
    response = speaker_client.post(confirmed_submission.urls.confirm, follow=True)
    confirmed_submission.refresh_from_db()
    assert response.status_code == 200
    assert confirmed_submission.state == SubmissionStates.CONFIRMED


@pytest.mark.django_db
def test_cannot_confirm_submitted_submission(speaker_client, submission):
    response = speaker_client.get(submission.urls.confirm, follow=True)
    submission.refresh_from_db()
    assert response.status_code == 200
    assert submission.state == SubmissionStates.SUBMITTED
    response = speaker_client.post(submission.urls.confirm, follow=True)
    submission.refresh_from_db()
    assert response.status_code == 200
    assert submission.state == SubmissionStates.SUBMITTED


@pytest.mark.django_db
def test_can_reconfirm_submission(speaker_client, accepted_submission):
    accepted_submission.state = SubmissionStates.CONFIRMED
    accepted_submission.save()
    response = speaker_client.get(accepted_submission.urls.confirm, follow=True)
    accepted_submission.refresh_from_db()
    assert response.status_code == 200
    assert accepted_submission.state == SubmissionStates.CONFIRMED


@pytest.mark.django_db
def test_cannot_confirm_rejected_submission(other_speaker_client, rejected_submission):
    rejected_submission.state = SubmissionStates.REJECTED
    rejected_submission.save()
    response = other_speaker_client.get(rejected_submission.urls.confirm, follow=True)
    rejected_submission.refresh_from_db()
    assert response.status_code == 200
    assert rejected_submission.state == SubmissionStates.REJECTED


@pytest.mark.django_db
def test_can_withdraw_submission(speaker_client, submission):
    response = speaker_client.get(submission.urls.withdraw, follow=True)
    submission.refresh_from_db()
    assert response.status_code == 200
    assert submission.state == SubmissionStates.SUBMITTED
    response = speaker_client.post(submission.urls.withdraw, follow=True)
    submission.refresh_from_db()
    assert response.status_code == 200
    assert submission.state == SubmissionStates.WITHDRAWN


@pytest.mark.django_db
def test_cannot_withdraw_accepted_submission(speaker_client, accepted_submission):
    response = speaker_client.get(accepted_submission.urls.withdraw, follow=True)
    accepted_submission.refresh_from_db()
    assert response.status_code == 200
    assert accepted_submission.state == SubmissionStates.ACCEPTED


@pytest.mark.django_db
def test_can_discard_draft_proposal(speaker_client, submission):
    with scope(event=submission.event):
        submission.state = SubmissionStates.DRAFT
        submission.save()
    response = speaker_client.get(submission.urls.discard, follow=True)
    assert response.status_code == 200
    response = speaker_client.post(submission.urls.discard, follow=True)
    assert response.status_code == 200
    with scope(event=submission.event):
        assert not submission.event.submissions.filter(pk=submission.pk).exists()


@pytest.mark.django_db
def test_cannot_discard_non_draft_proposal(speaker_client, submission):
    response = speaker_client.get(submission.urls.discard, follow=True)
    assert response.status_code == 404
    response = speaker_client.post(submission.urls.discard, follow=True)
    assert response.status_code == 404
    with scope(event=submission.event):
        assert submission.event.submissions.filter(pk=submission.pk).exists()


@pytest.mark.django_db
def test_can_edit_submission(speaker_client, submission, resource, other_resource):
    with scope(event=submission.event):
        assert submission.resources.count() == 2
        resource_one = submission.resources.first()
        resource_two = submission.resources.last()
        assert submission.title in str(resource_one)
        f = SimpleUploadedFile("testfile.txt", b"file_content")
        data = {
            "title": "Ein ganz neuer Titel",
            "submission_type": submission.submission_type.pk,
            "content_locale": submission.content_locale,
            "description": submission.description,
            "abstract": submission.abstract,
            "notes": submission.notes,
            "slot_count": submission.slot_count,
            "resource-0-id": resource_one.id,
            "resource-0-description": "new resource name",
            "resource-0-resource": resource_one.resource,
            "resource-1-id": resource_two.id,
            "resource-1-DELETE": True,
            "resource-1-description": resource_two.description,
            "resource-1-resource": resource_two.resource,
            "resource-2-id": "",
            "resource-2-description": "new resource",
            "resource-2-resource": f,
            "resource-TOTAL_FORMS": 3,
            "resource-INITIAL_FORMS": 2,
            "resource-MIN_NUM_FORMS": 0,
            "resource-MAX_NUM_FORMS": 1000,
        }
    response = speaker_client.post(submission.urls.user_base, follow=True, data=data)
    assert response.status_code == 200
    with scope(event=submission.event):
        assert submission.resources.count() == 2
        submission.refresh_from_db()
        resource_one.refresh_from_db()
        new_resource = submission.resources.exclude(pk=resource_one.pk).first()
        assert submission.title == "Ein ganz neuer Titel", response.content.decode()
        assert submission.resources.count() == 2
        assert new_resource.description == "new resource"
        assert new_resource.resource.read() == b"file_content"
        assert not submission.resources.filter(pk=resource_two.pk).exists()


@pytest.mark.django_db
def test_can_edit_slot_count(speaker_client, submission):
    with scope(event=submission.event):
        submission.event.feature_flags["present_multiple_times"] = True
        submission.event.save()
        data = {
            "title": "Ein ganz neuer Titel",
            "submission_type": submission.submission_type.pk,
            "content_locale": submission.content_locale,
            "description": submission.description,
            "abstract": submission.abstract,
            "notes": submission.notes,
            "slot_count": 13,
            "resource-TOTAL_FORMS": 0,
            "resource-INITIAL_FORMS": 0,
            "resource-MIN_NUM_FORMS": 0,
            "resource-MAX_NUM_FORMS": 1000,
        }
    response = speaker_client.post(submission.urls.user_base, follow=True, data=data)
    assert response.status_code == 200
    with scope(event=submission.event):
        submission.refresh_from_db()
        assert submission.slot_count == 13


@pytest.mark.django_db
def test_cannot_edit_confirmed_slot_count(speaker_client, confirmed_submission):
    submission = confirmed_submission
    submission.event.feature_flags["present_multiple_times"] = True
    submission.event.save()
    with scope(event=submission.event):
        data = {
            "title": "Ein ganz neuer Titel",
            "submission_type": submission.submission_type.pk,
            "content_locale": submission.content_locale,
            "description": submission.description,
            "abstract": submission.abstract,
            "notes": submission.notes,
            "slot_count": 13,
            "resource-TOTAL_FORMS": 0,
            "resource-INITIAL_FORMS": 0,
            "resource-MIN_NUM_FORMS": 0,
            "resource-MAX_NUM_FORMS": 1000,
        }
    response = speaker_client.post(submission.urls.user_base, follow=True, data=data)
    assert response.status_code == 200
    with scope(event=submission.event):
        submission.refresh_from_db()
        assert submission.slot_count != 13


@pytest.mark.django_db
def test_cannot_edit_rejected_submission(other_speaker_client, rejected_submission):
    title = rejected_submission.title
    data = {
        "title": "Ein ganz neuer Titel",
        "submission_type": rejected_submission.submission_type.pk,
        "content_locale": rejected_submission.content_locale,
        "description": rejected_submission.description,
        "abstract": rejected_submission.abstract,
        "notes": rejected_submission.notes,
        "resource-TOTAL_FORMS": 0,
        "resource-INITIAL_FORMS": 0,
        "resource-MIN_NUM_FORMS": 0,
        "resource-MAX_NUM_FORMS": 1000,
    }
    response = other_speaker_client.post(
        rejected_submission.urls.user_base, follow=True, data=data
    )
    assert response.status_code == 200
    rejected_submission.refresh_from_db()
    assert rejected_submission.title == title


@pytest.mark.django_db
def test_can_edit_submission_type(speaker_client, submission, event):
    with scope(event=submission.event):
        new_type = event.submission_types.create(name="Other", default_duration=13)
        data = {
            "title": "Ein ganz neuer Titel",
            "submission_type": new_type.pk,
            "content_locale": submission.content_locale,
            "description": submission.description,
            "abstract": submission.abstract,
            "notes": submission.notes,
            "resource-TOTAL_FORMS": 0,
            "resource-INITIAL_FORMS": 0,
            "resource-MIN_NUM_FORMS": 0,
            "resource-MAX_NUM_FORMS": 1000,
        }
    response = speaker_client.post(submission.urls.user_base, follow=True, data=data)
    assert response.status_code == 200
    with scope(event=submission.event):
        submission.refresh_from_db()
        assert submission.submission_type == new_type


@pytest.mark.django_db
def test_cannot_edit_submission_type_after_acceptance(
    speaker_client, submission, event
):
    with scope(event=submission.event):
        submission.accept()
        new_type = event.submission_types.create(name="Other", default_duration=13)
        data = {
            "title": "Ein ganz neuer Titel",
            "submission_type": new_type.pk,
            "content_locale": submission.content_locale,
            "description": submission.description,
            "abstract": submission.abstract,
            "notes": submission.notes,
            "resource-TOTAL_FORMS": 0,
            "resource-INITIAL_FORMS": 0,
            "resource-MIN_NUM_FORMS": 0,
            "resource-MAX_NUM_FORMS": 1000,
        }
    response = speaker_client.post(submission.urls.user_base, follow=True, data=data)
    assert response.status_code == 200
    with scope(event=submission.event):
        submission.refresh_from_db()
        assert submission.submission_type != new_type


@pytest.mark.django_db
def test_can_edit_profile(speaker, event, speaker_client):
    response = speaker_client.post(
        event.urls.user,
        data={
            "name": "Lady Imperator",
            "biography": "Ruling since forever.",
            "form": "profile",
        },
        follow=True,
    )
    assert response.status_code == 200
    with scope(event=event):
        speaker.refresh_from_db()
        assert speaker.profiles.get(event=event).biography == "Ruling since forever."
        assert speaker.name == "Lady Imperator"
    response = speaker_client.post(
        event.urls.user,
        data={
            "name": "Lady Imperator",
            "biography": "Ruling since forever.",
            "form": "profile",
        },
        follow=True,
    )
    assert response.status_code == 200
    with scope(event=event):
        speaker.refresh_from_db()
        assert speaker.profiles.get(event=event).biography == "Ruling since forever."
        assert speaker.name == "Lady Imperator"


@pytest.mark.django_db
def test_can_change_api_token(speaker, event, speaker_client):
    speaker.regenerate_token()
    old_token = Token.objects.filter(user=speaker).first().key
    response = speaker_client.post(
        event.urls.user,
        data={
            "form": "token",
        },
        follow=True,
    )
    assert response.status_code == 200
    new_token = Token.objects.filter(user=speaker).first().key
    assert new_token != old_token


@pytest.mark.django_db
def test_must_provide_availabilities(speaker, event, speaker_client):
    event.cfp.fields["availabilities"]["visibility"] = "required"
    event.cfp.save()
    response = speaker_client.post(
        event.urls.user,
        data={
            "name": "Lady Imperator",
            "biography": "Ruling since forever.",
            "form": "profile",
        },
        follow=True,
    )
    assert response.status_code == 200
    with scope(event=event):
        speaker.refresh_from_db()
        assert speaker.profiles.get(event=event).biography != "Ruling since forever."
    response = speaker_client.post(
        event.urls.user,
        data={
            "name": "Lady Imperator",
            "biography": "Ruling since forever.",
            "form": "profile",
            "availabilities": '{"availabilities": []}',
        },
        follow=True,
    )
    assert response.status_code == 200
    with scope(event=event):
        speaker.refresh_from_db()
        assert speaker.profiles.get(event=event).biography != "Ruling since forever."


@pytest.mark.django_db
def test_can_edit_login_info(speaker, event, speaker_client):
    response = speaker_client.post(
        event.urls.user,
        data={
            "old_password": "speakerpwd1!",
            "email": "new_email@speaker.org",
            "password": "",
            "password_repeat": "",
            "form": "login",
        },
        follow=True,
    )
    assert response.status_code == 200
    speaker.refresh_from_db()
    assert speaker.email == "new_email@speaker.org"


@pytest.mark.django_db
def test_can_edit_login_info_wrong_password(speaker, event, speaker_client):
    response = speaker_client.post(
        event.urls.user,
        data={
            "old_password": "speakerpwd23!",
            "email": "new_email@speaker.org",
            "password": "",
            "password_repeat": "",
            "form": "login",
        },
        follow=True,
    )
    assert response.status_code == 200
    speaker.refresh_from_db()
    assert speaker.email != "new_email@speaker.org"


@pytest.mark.django_db
def test_can_edit_and_update_speaker_answers(
    speaker,
    event,
    speaker_question,
    speaker_boolean_question,
    speaker_client,
    speaker_text_question,
    speaker_file_question,
):
    with scope(event=event):
        answer = speaker.answers.filter(question_id=speaker_question.pk).first()
        assert not answer
    f = SimpleUploadedFile("testfile.txt", b"file_content")
    response = speaker_client.post(
        event.urls.user,
        data={
            f"question_{speaker_question.id}": "black as the night",
            f"question_{speaker_boolean_question.id}": "True",
            f"question_{speaker_file_question.id}": f,
            f"question_{speaker_text_question.id}": "Green is totally the best color.",
            "form": "questions",
        },
        follow=True,
    )
    assert response.status_code == 200

    with scope(event=event):
        answer = speaker.answers.get(question_id=speaker_question.pk)
        assert answer.answer == "black as the night"
        assert (
            speaker.answers.get(question_id=speaker_boolean_question.pk).answer
            == "True"
        )
        assert (
            speaker.answers.get(question_id=speaker_text_question.pk).answer
            == "Green is totally the best color."
        )

        file_answer = speaker.answers.get(question_id=speaker_file_question.pk)
        assert file_answer.answer.startswith("file://")
        assert file_answer.answer_file.read() == b"file_content"
        assert (settings.MEDIA_ROOT / file_answer.answer_file.name).exists()

    response = speaker_client.post(
        event.urls.user,
        data={
            f"question_{speaker_question.id}": "green as the sky",
            "form": "questions",
        },
        follow=True,
    )
    assert response.status_code == 200
    with scope(event=event):
        answer.refresh_from_db()
        assert answer.answer == "green as the sky"

    response = speaker_client.post(
        event.urls.user,
        data={
            f"question_{speaker_question.id}": "green as the sky",
            "form": "questions",
        },
        follow=True,
    )
    assert response.status_code == 200
    with scope(event=event):
        answer.refresh_from_db()
        assert answer.answer == "green as the sky"


@pytest.mark.django_db
def test_cannot_delete_profile_on_first_try(speaker, event, speaker_client):
    with scope(event=event):
        assert speaker.profiles.get(event=event).biography != ""
    response = speaker_client.post(event.urls.user_delete, follow=True)
    assert response.status_code == 200
    with scope(event=event):
        speaker.refresh_from_db()
        assert speaker.profiles.get(event=event).biography != ""
        assert speaker.name != "Deleted User"


@pytest.mark.django_db
def test_can_delete_profile(speaker, event, speaker_client):
    with scope(event=event):
        assert speaker.profiles.get(event=event).biography != ""
    response = speaker_client.post(
        event.urls.user_delete, data={"really": True}, follow=True
    )
    assert response.status_code == 200
    with scope(event=event):
        speaker.refresh_from_db()
        assert speaker.profiles.get(event=event).biography == ""
        assert speaker.name == "Deleted User"
        assert speaker.email.startswith("deleted_user")
        assert speaker.email.endswith("@localhost")


@pytest.mark.django_db
def test_can_change_locale(multilingual_event, client):
    first_response = client.get(multilingual_event.cfp.urls.public, follow=True)
    assert "submission" in first_response.content.decode()
    assert "Einreichung" not in first_response.content.decode()
    second_response = client.get(
        reverse("cfp:locale.set", kwargs={"event": multilingual_event.slug})
        + f"?locale=de&next=/{multilingual_event.slug}/",
        follow=True,
    )
    assert "Einreichung" in second_response.content.decode()


@pytest.mark.django_db
def test_can_change_locale_with_queryparam(multilingual_event, client):
    first_response = client.get(multilingual_event.cfp.urls.public, follow=True)
    assert "submission" in first_response.content.decode()
    assert "Einreichung" not in first_response.content.decode()
    second_response = client.get(
        reverse("cfp:locale.set", kwargs={"event": multilingual_event.slug})
        + f"?locale=de&next=/{multilingual_event.slug}/?foo=bar",
        follow=True,
    )
    assert "Einreichung" in second_response.content.decode()


@pytest.mark.django_db
def test_persists_changed_locale(multilingual_event, orga_user, orga_client):
    assert orga_user.locale == "en"
    response = orga_client.get(
        reverse("cfp:locale.set", kwargs={"event": multilingual_event.slug})
        + f"?locale=de&next=/{multilingual_event.slug}/",
        follow=True,
    )
    orga_user.refresh_from_db()
    assert response.status_code == 200
    assert orga_user.locale == "de"


@pytest.mark.django_db
def test_can_invite_speaker(speaker_client, submission):
    djmail.outbox = []
    response = speaker_client.get(
        submission.urls.invite, follow=True, data={"email": "invalidemail"}
    )
    assert response.status_code == 200
    data = {
        "speaker": "other@speaker.org",
        "subject": "Please join!",
        "text": "C'mon, it will be fun!",
    }
    response = speaker_client.post(submission.urls.invite, follow=True, data=data)
    assert response.status_code == 200
    assert len(djmail.outbox) == 1
    assert djmail.outbox[0].to == ["other@speaker.org"]


@pytest.mark.django_db
def test_can_accept_invitation(orga_client, submission):
    assert submission.speakers.count() == 1
    response = orga_client.post(submission.urls.accept_invitation, follow=True)
    submission.refresh_from_db()
    assert response.status_code == 200
    assert submission.speakers.count() == 2


@pytest.mark.django_db
def test_wrong_acceptance_link(orga_client, submission):
    assert submission.speakers.count() == 1
    response = orga_client.post(
        submission.urls.accept_invitation + "olololol", follow=True
    )
    submission.refresh_from_db()
    assert response.status_code == 404
    assert submission.speakers.count() == 1


@pytest.mark.django_db
@pytest.mark.parametrize("get_availability", ("optional", "do_not_ask"))
def test_submission_accept(speaker_client, submission, get_availability):
    submission.event.cfp.fields["availabilities"]["visibility"] = get_availability
    submission.event.cfp.save()
    submission.state = SubmissionStates.ACCEPTED
    submission.save()

    response = speaker_client.post(submission.urls.confirm, follow=True)
    submission.refresh_from_db()

    assert response.status_code == 200
    assert submission.state == SubmissionStates.CONFIRMED


@pytest.mark.django_db
def test_submission_accept_with_missing_availability(speaker_client, submission):
    submission.event.cfp.fields["availabilities"]["visibility"] = "required"
    submission.event.cfp.save()
    submission.state = SubmissionStates.ACCEPTED
    submission.save()

    response = speaker_client.post(submission.urls.confirm, follow=True)
    submission.refresh_from_db()

    assert response.status_code == 200
    assert submission.state == SubmissionStates.ACCEPTED


@pytest.mark.django_db
def test_submission_accept_nologin(client, submission):
    submission.state = SubmissionStates.ACCEPTED
    submission.save()

    response = client.post(submission.urls.confirm, follow=True)
    submission.refresh_from_db()

    assert response.status_code == 200
    assert response.redirect_chain[-1][1] == 302
    assert "login/?next=" in response.redirect_chain[-1][0]
    assert submission.state == SubmissionStates.ACCEPTED


@pytest.mark.django_db
def test_submission_accept_wrong_code(client, submission):
    submission.state = SubmissionStates.ACCEPTED
    submission.save()

    assert submission.code in submission.urls.confirm
    response = client.post(
        submission.urls.confirm.replace(submission.code, "foo"), follow=True
    )

    assert response.status_code == 200
    assert response.redirect_chain[-1][1] == 302
    assert "login/?next=" in response.redirect_chain[-1][0]


@pytest.mark.django_db
def test_submission_withdraw(speaker_client, submission):
    djmail.outbox = []
    submission.state = SubmissionStates.SUBMITTED
    submission.save()

    response = speaker_client.post(submission.urls.withdraw, follow=True)
    assert response.status_code == 200
    submission.refresh_from_db()
    assert submission.state == SubmissionStates.WITHDRAWN
    assert len(djmail.outbox) == 0


@pytest.mark.django_db
def test_submission_withdraw_if_accepted(speaker_client, submission):
    djmail.outbox = []
    with scope(event=submission.event):
        submission.accept()

    response = speaker_client.post(submission.urls.withdraw, follow=True)
    assert response.status_code == 200
    with scope(event=submission.event):
        submission.refresh_from_db()
        assert submission.state == SubmissionStates.WITHDRAWN
        assert len(djmail.outbox) == 1


@pytest.mark.django_db
def test_submission_withdraw_if_confirmed(speaker_client, submission):
    with scope(event=submission.event):
        submission.accept()
        submission.confirm()

    response = speaker_client.post(submission.urls.withdraw, follow=True)
    assert response.status_code == 200
    with scope(event=submission.event):
        submission.refresh_from_db()
        assert submission.state != SubmissionStates.WITHDRAWN


@pytest.mark.django_db
def test_submission_withdraw_if_rejected(speaker_client, submission):
    with scope(event=submission.event):
        submission.reject()

    response = speaker_client.post(submission.urls.withdraw, follow=True)
    assert response.status_code == 200
    with scope(event=submission.event):
        submission.refresh_from_db()
        assert submission.state != SubmissionStates.WITHDRAWN



import json

import pytest
from django_scopes import scope

from pretalx.api.serializers.question import (
    AnswerSerializer,
    MinimalQuestionSerializer,
    QuestionSerializer,
)


@pytest.mark.django_db
def test_question_serializer(answer):
    with scope(event=answer.question.event):
        data = AnswerSerializer(answer).data
        assert set(data.keys()) == {
            "id",
            "question",
            "answer",
            "answer_file",
            "submission",
            "person",
            "review",
            "options",
        }
        data = QuestionSerializer(answer.question).data
    assert set(data.keys()) == {
        "id",
        "variant",
        "question",
        "question_required",
        "deadline",
        "freeze_after",
        "required",
        "read_only",
        "target",
        "options",
        "help_text",
        "default_answer",
        "contains_personal_data",
        "min_length",
        "max_length",
        "is_public",
        "is_visible_to_reviewers",
    }


@pytest.mark.django_db
def test_minimal_question_serializer(answer):
    with scope(event=answer.question.event):
        data = MinimalQuestionSerializer(answer.question).data
    assert set(data.keys()) == {"id", "question"}


@pytest.mark.django_db
@pytest.mark.parametrize("is_public", (True, False))
def test_questions_not_visible_by_default(client, question, schedule, is_public):
    with scope(event=question.event):
        question.is_public = is_public
        question.save()
    response = client.get(question.event.api_urls.questions, follow=True)
    content = json.loads(response.content.decode())
    assert response.status_code == 200
    assert bool(len(content["results"])) is is_public


@pytest.mark.django_db
def test_organizer_can_see_question(orga_client, question):
    response = orga_client.get(question.event.api_urls.questions, follow=True)
    content = json.loads(response.content.decode())
    assert response.status_code == 200
    assert len(content["results"]) == 1
    assert content["results"][0]["id"] == question.id


@pytest.mark.django_db
@pytest.mark.parametrize("is_visible", (True, False))
def test_questions_not_visible_by_default_to_reviewers(
    review_client, question, is_visible
):
    with scope(event=question.event):
        question.is_visible_to_reviewers = is_visible
        question.save()
    response = review_client.get(question.event.api_urls.questions, follow=True)
    content = json.loads(response.content.decode())
    assert response.status_code == 200
    assert bool(len(content["results"])) is is_visible


@pytest.mark.django_db
def test_organizer_can_create_question(event, orga_client):
    with scope(event=event):
        count = event.questions.all().count()
    response = orga_client.post(
        event.api_urls.questions,
        {
            "question": "A question",
            "variant": "text",
            "target": "submission",
            "help_text": "hellllp",
        },
    )
    assert response.status_code == 201, response.content.decode()
    with scope(event=event):
        assert event.questions.all().count() == count + 1
        question = event.questions.all().first()
        assert question.question == "A question"


@pytest.mark.django_db
def test_organizer_can_edit_question(event, orga_client, question):
    response = orga_client.patch(
        event.api_urls.questions + f"{question.pk}/",
        {
            "target": "speaker",
            "help_text": "hellllp",
        },
        content_type="application/json",
    )
    assert response.status_code == 200, response.content.decode()
    with scope(event=event):
        question.refresh_from_db()
        assert question.target == "speaker"
        assert question.help_text == "hellllp", response.content.decode()


@pytest.mark.django_db
def test_organizer_cannot_create_question_for_other_event(other_event, orga_client):
    with scope(event=other_event):
        count = other_event.questions.all().count()
    response = orga_client.post(
        other_event.api_urls.questions,
        {
            "question": "A question",
            "variant": "text",
            "target": "submission",
            "help_text": "hellllp",
        },
    )
    assert response.status_code == 403
    with scope(event=other_event):
        assert other_event.questions.all().count() == count


@pytest.mark.django_db
def test_reviewer_cannot_create_question(event, review_client):
    with scope(event=event):
        count = event.questions.all().count()
    response = review_client.post(
        event.api_urls.questions,
        {
            "question": "A question",
            "variant": "text",
            "target": "submission",
            "help_text": "hellllp",
        },
    )
    assert response.status_code == 403, response.content.decode()
    with scope(event=event):
        assert event.questions.all().count() == count


@pytest.mark.django_db
def test_reviewer_cannot_edit_question(event, review_client, question):
    response = review_client.patch(
        event.api_urls.questions + f"{question.pk}/",
        {
            "target": "speaker",
            "help_text": "hellllp",
        },
        content_type="application/json",
    )
    assert response.status_code == 403, response.content.decode()
    with scope(event=event):
        question.refresh_from_db()
        assert question.target != "speaker"
        assert question.help_text != "hellllp"



import datetime as dt
from urllib.parse import unquote

from django.conf import settings
from django.http import Http404, HttpResponse, JsonResponse
from django.utils.timezone import now
from django.views.decorators.cache import cache_page
from django.views.decorators.http import condition
from i18nfield.utils import I18nJSONEncoder

from pretalx.agenda.views.schedule import ScheduleView
from pretalx.common.tasks import generate_widget_css, generate_widget_js
from pretalx.common.utils import language
from pretalx.common.views import conditional_cache_page
from pretalx.schedule.exporters import ScheduleData


def widget_css_etag(request, **kwargs):
    return request.event.settings.widget_css_checksum


def widget_js_etag(request, event, version, locale="en", **kwargs):
    return request.event.settings.get(f"widget_checksum_{version}_{locale}")


def widget_data_etag(request, **kwargs):
    return request.event.settings.widget_data_checksum


class WidgetData(ScheduleView):
    def dispatch(self, request, *args, **kwargs):
        if not request.user.has_perm("agenda.view_widget", request.event):
            raise Http404()
        return super().dispatch(request, *args, **kwargs)

    def get_schedule_data_proportional(self, data):
        timezone = self.request.event.tz
        max_rooms = 0
        for date in data:
            if date.get("first_start") and date.get("last_end"):
                start = (
                    date.get("first_start")
                    .astimezone(timezone)
                    .replace(second=0, minute=0)
                )
                end = date.get("last_end").astimezone(timezone)
                height_seconds = (end - start).total_seconds()
                date["display_start"] = start
                date["height"] = int(height_seconds / 60 * 2)
                date["hours"] = []
                step = start
                while step < end:
                    date["hours"].append(step.strftime("%H:%M"))
                    step += dt.timedelta(hours=1)
                max_rooms = max(max_rooms, len(date["rooms"]))
                for room in date["rooms"]:
                    for talk in room.get("talks", []):
                        talk.top = int(
                            (talk.local_start - start).total_seconds() / 60 * 2
                        )
                        talk.height = int(talk.duration * 2)
                        talk.is_active = talk.start <= now() <= talk.real_end
        return {"data": list(data), "max_rooms": max_rooms}

    def get(self, request, *args, **kwargs):
        locale = request.GET.get("locale", "en")
        with language(locale):
            data = ScheduleData(
                event=self.request.event,
                schedule=self.schedule,
                with_accepted=False,
                with_breaks=True,
            ).data
            schedule = self.get_schedule_data_proportional(data)["data"]
            for day in schedule:
                for room in day["rooms"]:
                    room["name"] = str(room["name"])
                    room["talks"] = [
                        {
                            "title": talk.submission.title
                            if talk.submission
                            else str(talk.description),
                            "code": talk.submission.code if talk.submission else None,
                            "display_speaker_names": talk.submission.display_speaker_names
                            if talk.submission
                            else None,
                            "speakers": [
                                {"name": speaker.name, "code": speaker.code}
                                for speaker in talk.submission.speakers.all()
                            ]
                            if talk.submission
                            else None,
                            "height": talk.height,
                            "top": talk.top,
                            "start": talk.start,
                            "end": talk.end,
                            "do_not_record": talk.submission.do_not_record
                            if talk.submission
                            else None,
                            "track": getattr(talk.submission.track, "name", "")
                            if talk.submission
                            else None,
                            "featured": talk.submission.is_featured,
                        }
                        for talk in room["talks"]
                    ]
            response = JsonResponse(
                {
                    "schedule": schedule,
                    "event": {
                        "url": request.event.urls.schedule.full(),
                        "tracks": [
                            {"name": track.name, "color": track.color}
                            for track in request.event.tracks.all()
                        ],
                    },
                },
                encoder=I18nJSONEncoder,
            )
            response["Access-Control-Allow-Origin"] = "*"
            return response


def cache_control(request, event, version=None):
    """We don't want cache headers on unversioned and WIP schedule sites.

    This differs from where we actually cache: We cache unversioned
    sites, but we don't want clients to know about it, to make sure
    they'll get the most recent content upon cache invalidation.
    """
    if version:
        if version == "wip":
            return False
        return True
    return False


def cache_version(request, event, version=None):
    """We want to cache all pages except for the WIP schedule site.

    Note that unversioned pages will be cached, but no cache headers
    will be sent to make sure users will always see the most recent
    changes.
    """
    if version and version == "wip":
        return False
    return True


def version_prefix(request, event, version=None):
    """On non-versioned pages, we want cache-invalidation on schedule
    relese."""
    if not version and request.event.current_schedule:
        return request.event.current_schedule.version


@conditional_cache_page(
    60, cache_version, cache_control=cache_control, key_prefix=version_prefix
)
def widget_data_v2(request, event, version=None):
    event = request.event
    if request.method == "OPTIONS":
        response = JsonResponse({})
        response["Access-Control-Allow-Origin"] = "*"
        response["Access-Control-Allow-Headers"] = "authorization,content-type"
        return response
    if not request.user.has_perm("agenda.view_widget", event):
        raise Http404()

    version = version or unquote(request.GET.get("v") or "")
    schedule = None
    if version and version == "wip":
        if not request.user.has_perm("orga.view_schedule", event):
            raise Http404()
        schedule = request.event.wip_schedule
    elif version:
        schedule = event.schedules.filter(version__iexact=version).first()

    schedule = schedule or event.current_schedule
    if not schedule:
        raise Http404()

    result = schedule.build_data(all_talks=not schedule.version)
    response = JsonResponse(result, encoder=I18nJSONEncoder)
    response["Access-Control-Allow-Headers"] = "authorization,content-type"
    response["Access-Control-Allow-Origin"] = "*"
    return response


@condition(etag_func=widget_js_etag)
@cache_page(60)
def widget_script(request, event, locale=None, version=2):
    """The locale parameter is only relevant to the deprecated v1 version of
    the widget."""
    if not request.user.has_perm("agenda.view_widget", request.event):
        raise Http404()
    if locale and locale not in [lc[:2] for lc, ll in settings.LANGUAGES]:
        raise Http404()

    fname = f"widget_file_{version}"
    if locale:
        fname = "{fname}_{locale}"
    existing_file = request.event.settings.get(fname)
    if existing_file and not settings.DEBUG:  # pragma: no cover
        return HttpResponse(existing_file.read(), content_type="text/javascript")

    data = generate_widget_js(
        request.event, locale=locale, save=not settings.DEBUG, version=version
    )
    return HttpResponse(data, content_type="text/javascript")


@condition(etag_func=widget_css_etag)
@cache_page(60)
def widget_style(request, event, version):
    if not request.user.has_perm("agenda.view_widget", request.event):
        raise Http404()
    existing_file = request.event.settings.widget_css
    if existing_file and not settings.DEBUG:  # pragma: no cover
        return HttpResponse(existing_file.read(), content_type="text/css")

    data = generate_widget_css(request.event, save=not settings.DEBUG)
    return HttpResponse(data, content_type="text/css")



import json
from functools import partial

from django import forms
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_scopes.forms import SafeModelMultipleChoiceField

from pretalx.common.forms.widgets import MarkdownWidget
from pretalx.common.mixins.forms import ReadOnlyFlag
from pretalx.common.phrases import phrases
from pretalx.orga.forms.export import ExportForm
from pretalx.person.models import User
from pretalx.submission.models import Question, Review, Submission


class TagsForm(ReadOnlyFlag, forms.ModelForm):
    def __init__(self, event, **kwargs):
        self.event = event
        super().__init__(**kwargs)
        if not self.event.tags.all().exists():
            self.fields.pop("tags")
        else:
            self.fields["tags"].queryset = self.event.tags.all()
            self.fields["tags"].required = False

    class Meta:
        model = Submission
        fields = [
            "tags",
        ]
        widgets = {
            "tags": forms.SelectMultiple(attrs={"class": "select2"}),
        }
        field_classes = {
            "tags": SafeModelMultipleChoiceField,
        }


class ReviewForm(ReadOnlyFlag, forms.ModelForm):
    def __init__(self, event, user, *args, instance=None, categories=None, **kwargs):
        self.event = event
        self.categories = categories

        super().__init__(*args, instance=instance, **kwargs)

        # We validate existing score/text server-side to allow form-submit to skip/abstain
        self.fields["text"].required = False
        if self.event.review_settings["text_mandatory"]:
            self.fields["text"].widget.attrs["class"] = "hide-optional"

        self.scores = (
            {
                score.category: score.id
                for score in self.instance.scores.all().select_related("category")
            }
            if self.instance.id
            else {}
        )
        for category in categories:
            self.fields[f"score_{category.id}"] = self.build_score_field(
                category,
                read_only=kwargs.get("read_only", False),
                initial=self.scores.get(category),
                hide_optional=self.event.review_settings["score_mandatory"],
            )
            setattr(
                self,
                f"clean_score_{category.id}",
                partial(self._clean_score_category, category),
            )
        self.fields["text"].widget.attrs["rows"] = 2
        self.fields["text"].widget.attrs["placeholder"] = phrases.orga.example_review
        self.fields["text"].help_text += " " + phrases.base.use_markdown

    def build_score_field(
        self, category, read_only=False, initial=None, hide_optional=False
    ):
        choices = [(None, _("No score"))] if not category.required else []
        for score in category.scores.all():
            choices.append(
                (
                    score.id,
                    score.format(
                        self.event.review_settings.get("score_format", "words_numbers")
                    ),
                )
            )

        field = forms.ChoiceField(
            choices=choices,
            required=False,
            widget=forms.RadioSelect,
            disabled=read_only,
            initial=initial,
            label=category.name,
        )
        field.widget.attrs["autocomplete"] = "off"
        if hide_optional:
            field.widget.attrs["class"] = "hide-optional"
        return field

    def get_score_fields(self):
        for category in self.categories:
            yield self[f"score_{category.id}"]

    def clean_text(self):
        text = self.cleaned_data.get("text")
        if not text and self.event.review_settings["text_mandatory"]:
            raise forms.ValidationError(_("Please provide a review text!"))
        return text

    def _clean_score_category(self, category):
        score = self.cleaned_data.get(f"score_{category.id}")
        if score in ("", None) and category.required:
            raise forms.ValidationError(_("Please provide a review score!"))
        return score

    def save(self, *args, **kwargs):
        instance = super().save(*args, **kwargs)
        current_scores = []
        for category in self.categories:
            score_id = self.cleaned_data.get(f"score_{category.id}")
            if score_id:
                current_scores.append(score_id)
        instance.scores.set(current_scores)
        instance.save()
        return instance

    class Meta:
        model = Review
        fields = ("text",)
        widgets = {
            "text": MarkdownWidget,
        }


class DirectionForm(forms.Form):
    direction = forms.ChoiceField(
        choices=(
            ("reviewer", _("Assign proposals to reviewers")),
            ("submission", _("Assign reviewers to proposals")),
        ),
        required=False,
    )


class ReviewerForProposalForm(forms.ModelForm):
    def __init__(self, *args, reviewers=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["assigned_reviewers"].queryset = reviewers
        self.fields["assigned_reviewers"].label = self.instance.title

    def save(self, *args, **kwargs):
        # No calling 'super().save()' here – it would potentially update a submission's code!
        instance = self.instance
        if "assigned_reviewers" in self.changed_data:
            new_code = self.cleaned_data.get("code")
            if instance.code != new_code:
                instance = instance.event.submissions.all().get(code=new_code)
            instance.assigned_reviewers.set(self.cleaned_data["assigned_reviewers"])

    class Meta:
        model = Submission
        fields = ["assigned_reviewers", "code"]
        widgets = {
            "assigned_reviewers": forms.SelectMultiple(attrs={"class": "select2"}),
            "code": forms.HiddenInput(),
        }


class ProposalForReviewerForm(forms.ModelForm):
    def __init__(self, *args, proposals=None, **kwargs):
        super().__init__(*args, **kwargs)
        initial = proposals.filter(assigned_reviewers__in=[self.instance]).values_list(
            "id", flat=True
        )
        self.fields["assigned_reviews"] = forms.MultipleChoiceField(
            choices=((p.id, p.title) for p in proposals),
            widget=forms.SelectMultiple(attrs={"class": "select2"}),
            label=self.instance.name,
            initial=list(initial),
            required=False,
        )

    def save(self, *args, **kwargs):
        # No calling 'super().save()' here – it would potentially update a user's code!
        instance = self.instance
        if "assigned_reviews" in self.changed_data:
            new_code = self.cleaned_data.get("code")
            if instance.code != new_code:
                instance = User.objects.get(code=new_code)
            instance.assigned_reviews.set(self.cleaned_data["assigned_reviews"])

    class Meta:
        model = User
        fields = ["code"]
        widgets = {"code": forms.HiddenInput()}


class ReviewExportForm(ExportForm):
    data_delimiter = None
    target = forms.ChoiceField(
        required=True,
        label=_("Proposal"),
        choices=(
            ("all", _("All proposals")),
            ("accepted", _("accepted")),
            ("confirmed", _("confirmed")),
            ("rejected", _("rejected")),
        ),
        widget=forms.RadioSelect,
    )
    submission_id = forms.BooleanField(
        required=False,
        label=_("Proposal ID"),
        help_text=_(
            "The unique ID of a proposal is used in the proposal URL and in exports"
        ),
    )
    submission_title = forms.BooleanField(
        required=False,
        label=_("Proposal title"),
    )
    user_name = forms.BooleanField(
        required=False,
        label=_("Reviewer name"),
    )
    user_email = forms.BooleanField(
        required=False,
        label=_("Reviewer email"),
    )

    class Meta:
        model = Review
        model_fields = ["score", "text", "created", "updated"]

    def __init__(self, *args, user=None, **kwargs):
        self.user = user
        super().__init__(*args, **kwargs)
        self.fields["text"].label = _("Text")
        self._build_score_fields()

    @cached_property
    def questions(self):
        return Question.all_objects.filter(
            target="reviewer",
            active=True,
            event=self.event,
        ).prefetch_related("answers", "answers__review", "options")

    @cached_property
    def score_categories(self):
        sc = self.event.score_categories.filter(active=True)
        if len(sc) == 1:
            return []
        return sc

    @cached_property
    def score_field_names(self):
        return [f"score_{sc.pk}" for sc in self.score_categories]

    @cached_property
    def filename(self):
        return f"{self.event.slug}_reviews"

    @cached_property
    def export_field_names(self):
        return (
            [
                "score",
                "text",
            ]
            + self.score_field_names
            + [
                "submission_id",
                "submission_title",
                "created",
                "updated",
                "user_name",
                "user_email",
            ]
        )

    def _build_score_fields(self):
        for score_category in self.score_categories:
            self.fields[f"score_{score_category.pk}"] = forms.BooleanField(
                required=False,
                label=str(_("Score in '{score_category}'")).format(
                    score_category=score_category.name
                ),
            )

    def get_additional_data(self, obj):
        return {
            str(_("Score in '{score_category}'")).format(
                score_category=sc.name
            ): getattr(obj.scores.filter(category=sc).first(), "value", None)
            for sc in self.score_categories
        }

    def get_queryset(self):
        target = self.cleaned_data.get("target")
        queryset = Review.objects.filter(submission__event=self.event)
        if target != "all":
            queryset = queryset.filter(
                submission__in=self.event.submissions.filter(state=target)
            ).distinct()
        # TODO auto-adjust further to available tracks etc
        queryset = queryset.exclude(submission__speakers__in=[self.user]).distinct()
        return queryset.select_related("submission", "user").prefetch_related(
            "answers", "answers__question", "scores", "scores__category"
        )

    def _get_submission_id_value(self, obj):
        return obj.submission.code

    def _get_submission_title_value(self, obj):
        return obj.submission.title

    def _get_user_name_value(self, obj):
        return obj.user.name

    def _get_user_email_value(self, obj):
        return obj.user.email

    def get_answer(self, question, obj):
        return question.answers.filter(review=obj).first()


class ReviewAssignImportForm(DirectionForm):
    import_file = forms.FileField(label=_("file"))
    replace_assignments = forms.ChoiceField(
        label=_("Replace current assignments"),
        choices=((0, _("No")), (1, _("Yes"))),
        help_text=_(
            "Select to remove all current assignments and replace them with the import. Otherwise, the import will be an addition to the current assignments."
        ),
        widget=forms.RadioSelect,
        initial=False,
    )

    def __init__(self, event, **kwargs):
        self.event = event
        self._user_cache = {}
        self._submissions_cache = {}
        super().__init__(**kwargs)
        self.fields["direction"].required = True

    def _get_user(self, text):
        if text in self._user_cache:
            return self._user_cache[text]
        try:
            user = self.event.reviewers.get(Q(email__iexact=text) | Q(code=text))
            self._user_cache[text] = user
            return user
        except Exception:
            raise forms.ValidationError(str(_("Unknown user: {}")).format(text))

    def _get_submission(self, text):
        if not self._submissions_cache:
            self._submissions_cache = {s.code: s for s in self.event.submissions.all()}
        try:
            return self._submissions_cache[text.strip().upper()]
        except Exception:
            raise forms.ValidationError(str(_("Unknown proposal: {}")).format(text))

    def clean_import_file(self):
        uploaded_file = self.cleaned_data["import_file"]
        try:
            data = json.load(uploaded_file)
        except Exception:
            raise forms.ValidationError(_("Cannot parse JSON file."))
        return data

    def clean(self):
        super().clean()
        uploaded_data = self.cleaned_data.get("import_file")
        direction = self.cleaned_data.get("direction")
        if not uploaded_data:
            raise forms.ValidationError(_("Cannot parse JSON file."))
        if direction == "reviewer":
            # keys should be users, values should be lists of proposals
            new_uploaded_data = {
                self._get_user(key): [self._get_submission(v) for v in value]
                for key, value in uploaded_data.items()
            }
        else:
            # keys should be proposals, values should be lists of users
            new_uploaded_data = {
                self._get_submission(key): [self._get_user(v) for v in value]
                for key, value in uploaded_data.items()
            }

        self.cleaned_data["import_file"] = new_uploaded_data
        return self.cleaned_data

    def save(self):
        direction = self.cleaned_data.get("direction")
        replace_assignments = self.cleaned_data.get("replace_assignments")
        uploaded_data = self.cleaned_data.get("import_file")

        if replace_assignments in (1, "1"):
            # There's no .update() for m2m fields
            # We'll just assume that there are less reviewers than proposals, typically,
            # so this should result in less queries
            for user in self.event.reviewers:
                user.assigned_reviews.set([])

        if direction == "reviewer":
            # keys should be users, values should be lists of proposals
            for user, proposals in uploaded_data.items():
                user.assigned_reviews.add(*proposals)
        else:
            for proposal, users in uploaded_data.items():
                proposal.assigned_reviewers.add(*users)



import pytest

from pretalx.common.forms.renderers import render_label


@pytest.mark.parametrize(
    "label_class,label_title,expected",
    (
        (None, None, '<label for="forrr">content</label>'),
        ("clllass", None, '<label class="clllass" for="forrr">content</label>'),
        (None, "tiiitle", '<label for="forrr" title="tiiitle">content</label>'),
        (
            "clllass",
            "tiiitle",
            '<label class="clllass" for="forrr" title="tiiitle">content</label>',
        ),
    ),
)
def test_render(label_class, label_title, expected):
    assert (
        render_label(
            "content",
            label_for="forrr",
            label_class=label_class,
            label_title=label_title,
        )
        == expected
    )
